"""
Local backend

Execute the flow with a native runtime
using local / remote processes
"""
from __future__ import print_function
import json
import os
import sys
import fcntl
import time
import select
import subprocess
from datetime import datetime
from io import BytesIO
from functools import partial

from . import get_namespace
from .metaflow_config import MAX_ATTEMPTS
from .exception import MetaflowException,\
    MetaflowInternalError,\
    METAFLOW_EXIT_DISALLOW_RETRY
from . import procpoll
from .datastore import DataException, MetaflowDatastoreSet
from .metadata import MetaDatum
from .debug import debug
from .decorators import flow_decorators
from .mflog import mflog, RUNTIME_LOG_SOURCE
from .util import to_unicode, compress_list, unicode_type
from .unbounded_foreach import CONTROL_TASK_TAG, UBF_CONTROL, UBF_TASK

MAX_WORKERS=16
MAX_NUM_SPLITS=100
MAX_LOG_SIZE=1024*1024
PROGRESS_INTERVAL = 1000 #ms
# The following is a list of the (data) artifacts used by the runtime while
# executing a flow. These are prefetched during the resume operation by
# leveraging the MetaflowDatastoreSet.
PREFETCH_DATA_ARTIFACTS = ['_foreach_stack', '_task_ok', '_transition']

# Runtime must use logsource=RUNTIME_LOG_SOURCE for all loglines that it
# formats according to mflog. See a comment in mflog.__init__
mflog_msg = partial(mflog.decorate, RUNTIME_LOG_SOURCE)

# TODO option: output dot graph periodically about execution

class NativeRuntime(object):

    def __init__(self,
                 flow,
                 graph,
                 datastore,
                 metadata,
                 environment,
                 package,
                 logger,
                 entrypoint,
                 event_logger,
                 monitor,
                 run_id=None,
                 clone_run_id=None,
                 clone_steps=None,
                 max_workers=MAX_WORKERS,
                 max_num_splits=MAX_NUM_SPLITS,
                 max_log_size=MAX_LOG_SIZE):

        if run_id is None:
            self._run_id = metadata.new_run_id()
        else:
            self._run_id = run_id
            metadata.register_run_id(run_id)

        self._flow = flow
        self._graph = graph
        self._datastore = datastore
        self._metadata = metadata
        self._environment = environment
        self._logger = logger
        self._max_workers = max_workers
        self._num_active_workers = 0
        self._max_num_splits = max_num_splits
        self._max_log_size = max_log_size
        self._params_task = None
        self._entrypoint = entrypoint
        self.event_logger = event_logger
        self._monitor = monitor

        self._clone_run_id = clone_run_id
        self._clone_steps = {} if clone_steps is None else clone_steps

        self._origin_ds_set = None
        if clone_run_id:
            # resume logic
            # 0. If clone_run_id is specified, attempt to clone all the
            # successful tasks from the flow with `clone_run_id`. And run the
            # unsuccessful or not-run steps in the regular fashion.
            # 1. With _find_origin_task, for every task in the current run, we
            # find the equivalent task in `clone_run_id` using
            # pathspec_index=run/step:[index] and verify if this task can be
            # cloned.
            # 2. If yes, we fire off a clone-only task which copies the
            # metadata from the `clone_origin` to pathspec=run/step/task to
            # mimmick that the resumed run looks like an actual run.
            # 3. All steps that couldn't be cloned (either unsuccessful or not
            # run) are run as regular tasks.
            # Lastly, to improve the performance of the cloning process, we
            # leverage the MetaflowDatastoreSet abstraction to prefetch the
            # entire DAG of `clone_run_id` and relevant data artifacts
            # (see PREFETCH_DATA_ARTIFACTS) so that the entire runtime can
            # access the relevant data from cache (instead of going to the datastore
            # after the first prefetch).
            logger(
                'Gathering required information to resume run (this may take a bit of time)...')
            self._origin_ds_set = \
                MetaflowDatastoreSet(
                    datastore,
                    flow.name,
                    clone_run_id,
                    metadata=metadata,
                    event_logger=event_logger,
                    monitor=monitor,
                    prefetch_data_artifacts=PREFETCH_DATA_ARTIFACTS)
        self._run_queue = []
        self._poll = procpoll.make_poll()
        self._workers = {}  # fd -> subprocess mapping
        self._finished = {}
        self._is_cloned = {}
        # NOTE: In case of unbounded foreach, we need the following to schedule
        # the (sibling) mapper tasks  of the control task (in case of resume);
        # and ensure that the join tasks runs only if all dependent tasks have
        # finished.
        self._control_num_splits = {}  # control_task -> num_splits mapping

        for step in flow:
            for deco in step.decorators:
                deco.runtime_init(flow,
                                  graph,
                                  package,
                                  self._run_id)

    def _new_task(self, step, input_paths=None, **kwargs):

        if input_paths is None:
            may_clone = True
        else:
            may_clone = all(self._is_cloned[path] for path in input_paths)

        if step in self._clone_steps:
            may_clone = False

        if step == '_parameters':
            decos = []
        else:
            decos = getattr(self._flow, step).decorators

        return Task(self._datastore,
                    self._flow,
                    step,
                    self._run_id,
                    self._metadata,
                    self._environment,
                    self._entrypoint,
                    self.event_logger,
                    self._monitor,
                    input_paths=input_paths,
                    may_clone=may_clone,
                    clone_run_id=self._clone_run_id,
                    origin_ds_set=self._origin_ds_set,
                    decos=decos,
                    logger=self._logger,
                    **kwargs)

    @property
    def run_id(self):
        return self._run_id

    def persist_parameters(self, task_id=None):
        task = self._new_task('_parameters', task_id=task_id)
        if not task.is_cloned:
            task.persist(self._flow)
        self._params_task = task.path
        self._is_cloned[task.path] = task.is_cloned

    def execute(self):

        self._logger('Workflow starting (run-id %s):' % self._run_id,
                     system_msg=True)

        self._metadata.start_run_heartbeat(self._flow.name, self._run_id)

        if self._params_task:
            self._queue_push('start', {'input_paths': [self._params_task]})
        else:
            self._queue_push('start', {})

        progress_tstamp = time.time()
        try:
            # main scheduling loop
            exception = None
            while self._run_queue or\
                    self._num_active_workers > 0:

                # 1. are any of the current workers finished?
                finished_tasks = list(self._poll_workers())
                # 2. push new tasks triggered by the finished tasks to the queue
                self._queue_tasks(finished_tasks)
                # 3. if there are available worker slots, pop and start tasks
                #    from the queue.
                self._launch_workers()

                if time.time() - progress_tstamp > PROGRESS_INTERVAL:
                    progress_tstamp = time.time()
                    msg = "%d tasks are running: %s." %\
                          (self._num_active_workers, 'e.g. ...')  # TODO
                    self._logger(msg, system_msg=True)
                    msg = "%d tasks are waiting in the queue." %\
                          len(self._run_queue)
                    self._logger(msg, system_msg=True)
                    msg = "%d steps are pending: %s." %\
                          (0, 'e.g. ...')  # TODO
                    self._logger(msg, system_msg=True)

        except KeyboardInterrupt as ex:
            self._logger('Workflow interrupted.', system_msg=True, bad=True)
            self._killall()
            exception = ex
            raise
        except Exception as ex:
            self._logger('Workflow failed.', system_msg=True, bad=True)
            self._killall()
            exception = ex
            raise
        finally:
            # on finish clean tasks
            for step in self._flow:
                for deco in step.decorators:
                    deco.runtime_finished(exception)

            self._metadata.stop_heartbeat()

        # assert that end was executed and it was successful
        if ('end', ()) in self._finished:
            self._logger('Done!', system_msg=True)
        else:
            raise MetaflowInternalError('The *end* step was not successful '
                                        'by the end of flow.')

    def _killall(self):
        # If we are here, all children have received a signal and are shutting down.
        # We want to give them an opportunity to do so and then kill
        live_workers = set(self._workers.values())
        now = int(time.time())
        self._logger('Terminating %d active tasks...' % len(live_workers),
                     system_msg=True, bad=True)
        while live_workers and int(time.time()) - now < 5:
            # While not all workers are dead and we have waited less than 5 seconds
            live_workers = [worker for worker in live_workers if not worker.clean()]
        if live_workers:
            self._logger('Killing %d remaining tasks after having waited for %d seconds -- '
                         'some tasks may not exit cleanly' % (len(live_workers),
                                                              int(time.time()) - now),
                         system_msg=True, bad=True)
            for worker in live_workers:
                worker.kill()
        self._logger('Flushing logs...', system_msg=True, bad=True)
        # give killed workers a chance to flush their logs to datastore
        for _ in range(3):
            list(self._poll_workers())

    # Store the parameters needed for task creation, so that pushing on items
    # onto the run_queue is an inexpensive operation.
    def _queue_push(self, step, task_kwargs):
        self._run_queue.insert(0, (step, task_kwargs))

    def _queue_pop(self):
        return self._run_queue.pop() if self._run_queue else None

    def _queue_task_join(self, task, next_steps):
        # if the next step is a join, we need to check that
        # all input tasks for the join have finished before queuing it.

        # CHECK: this condition should be enforced by the linter but
        # let's assert that the assumption holds
        if len(next_steps) > 1:
            msg = 'Step *{step}* transitions to a join and another '\
                  'step. The join must be the only transition.'
            raise MetaflowInternalError(task, msg.format(step=task.step))
        else:
            next_step = next_steps[0]


        unbounded_foreach = not task.results.is_none('_unbounded_foreach')

        if unbounded_foreach:
            # Before we queue the join, do some post-processing of runtime state
            # (_finished, _is_cloned) for the (sibling) mapper tasks.
            # Update state of (sibling) mapper tasks for control task.
            if task.ubf_context == UBF_CONTROL:
                mapper_tasks = task.results.get('_control_mapper_tasks')
                if not mapper_tasks:
                    msg = "Step *{step}* has a control task which didn't "\
                          "specify the artifact *_control_mapper_tasks* for "\
                          "the subsequent *{join}* step."
                    raise MetaflowInternalError(msg.format(step=task.step,
                                                           join=next_steps[0]))
                elif not (isinstance(mapper_tasks, list) and\
                          isinstance(mapper_tasks[0], unicode_type)):
                    msg = "Step *{step}* has a control task which didn't "\
                          "specify the artifact *_control_mapper_tasks* as a "\
                          "list of strings but instead specified it as {typ} "\
                          "with elements of {elem_typ}."
                    raise MetaflowInternalError(
                            msg.format(step=task.step,
                                       typ=type(mapper_tasks),
                                       elem_type=type(mapper_tasks[0])))
                num_splits = len(mapper_tasks)
                self._control_num_splits[task.path] = num_splits
                if task.is_cloned:
                    # Add mapper tasks to be cloned.
                    for i in range(num_splits):
                        # NOTE: For improved robustness, introduce 
                        # `clone_options` as an enum so that we can force that 
                        # clone must occur for this task.
                        self._queue_push(task.step,
                                         {'input_paths': task.input_paths,
                                          'split_index': str(i),
                                          'ubf_context': UBF_TASK})
                else:
                    # Update _finished since these tasks were successfully
                    # run elsewhere so that join will be unblocked.
                    step_name, foreach_stack = task.finished_id
                    top = foreach_stack[-1]
                    bottom = list(foreach_stack[:-1])
                    for i in range(num_splits):
                        s = tuple(bottom + [top._replace(index=i)])
                        self._finished[(task.step, s)] = mapper_tasks[i]
                        self._is_cloned[mapper_tasks[i]] = False

            # Find and check status of control task and retrieve its pathspec
            # for retrieving unbounded foreach cardinality.
            step_name, foreach_stack = task.finished_id
            top = foreach_stack[-1]
            bottom = list(foreach_stack[:-1])
            s = tuple(bottom + [top._replace(index=None)])
            control_path = self._finished.get((task.step, s))
            if control_path:
                # Control task was successful.
                # Additionally check the state of (sibling) mapper tasks as well
                # (for the sake of resume) before queueing join task.
                num_splits = self._control_num_splits[control_path]
                required_tasks = []
                for i in range(num_splits):
                    s = tuple(bottom + [top._replace(index=i)])
                    required_tasks.append(self._finished.get((task.step, s)))
                required_tasks.append(control_path)

                if all(required_tasks):
                    # all tasks to be joined are ready. Schedule the next join step.
                    self._queue_push(next_step,
                                     {'input_paths': required_tasks,
                                      'join_type': 'foreach'})
        else:
            # matching_split is the split-parent of the finished task
            matching_split = \
                        self._graph[self._graph[next_step].split_parents[-1]]
            step_name, foreach_stack = task.finished_id

            if matching_split.type == 'foreach':
                # next step is a foreach join

                def siblings(foreach_stack):
                    top = foreach_stack[-1]
                    bottom = list(foreach_stack[:-1])
                    for index in range(top.num_splits):
                        yield tuple(bottom + [top._replace(index=index)])

                # required tasks are all split-siblings of the finished task
                required_tasks = [self._finished.get((task.step, s))
                                  for s in siblings(foreach_stack)]
                join_type = 'foreach'
            else:
                # next step is a split-and
                # required tasks are all branches joined by the next step
                required_tasks = [self._finished.get((step, foreach_stack))
                                  for step in self._graph[next_step].in_funcs]
                join_type = 'linear'

            if all(required_tasks):
                # all tasks to be joined are ready. Schedule the next join step.
                self._queue_push(next_step,
                                 {'input_paths': required_tasks,
                                  'join_type': join_type})

    def _queue_task_foreach(self, task, next_steps):

        # CHECK: this condition should be enforced by the linter but
        # let's assert that the assumption holds
        if len(next_steps) > 1:
            msg = 'Step *{step}* makes a foreach split but it defines '\
                  'multiple transitions. Specify only one transition '\
                  'for foreach.'
            raise MetaflowInternalError(msg.format(step=task.step))
        else:
            next_step = next_steps[0]

        unbounded_foreach = not task.results.is_none('_unbounded_foreach')

        if unbounded_foreach:
            # Need to push control process related task.
            self._queue_push(next_step,
                             {'input_paths': [task.path],
                              'ubf_context': UBF_CONTROL})  
        else:
            num_splits = task.results['_foreach_num_splits']
            if num_splits > self._max_num_splits:
                msg = 'Foreach in step *{step}* yielded {num} child steps '\
                      'which is more than the current maximum of {max} '\
                      'children. You can raise the maximum with the '\
                      '--max-num-splits option. '
                raise TaskFailed(task, msg.format(step=task.step,
                                                  num=num_splits,
                                                  max=self._max_num_splits))

            # schedule all splits
            for i in range(num_splits):
                self._queue_push(next_step,
                                 {'split_index': str(i),
                                  'input_paths': [task.path]})

    def _queue_tasks(self, finished_tasks):
        # finished tasks include only successful tasks
        for task in finished_tasks:
            self._finished[task.finished_id] = task.path
            self._is_cloned[task.path] = task.is_cloned

            # CHECK: ensure that runtime transitions match with
            # statically inferred transitions. Make an exception for control
            # tasks, where we just rely on static analysis since we don't
            # execute user code.
            trans = task.results.get('_transition')
            if trans:
                next_steps = trans[0]
                foreach = trans[1]
            else:
                next_steps = []
                foreach = None
            expected = self._graph[task.step].out_funcs
            if next_steps != expected:
                msg = 'Based on static analysis of the code, step *{step}* '\
                      'was expected to transition to step(s) *{expected}*. '\
                      'However, when the code was executed, self.next() was '\
                      'called with *{actual}*. Make sure there is only one '\
                      'unconditional self.next() call in the end of your '\
                      'step. '
                raise MetaflowInternalError(msg.format(step=task.step,
                                                       expected=', '.join(
                                                           expected),
                                                       actual=', '.join(next_steps)))

            # Different transition types require different treatment
            if any(self._graph[f].type == 'join' for f in next_steps):
                # Next step is a join
                self._queue_task_join(task, next_steps)
            elif foreach:
                # Next step is a foreach child
                self._queue_task_foreach(task, next_steps)
            else:
                # Next steps are normal linear steps
                for step in next_steps:
                    self._queue_push(step, {'input_paths': [task.path]})

    def _poll_workers(self):
        if self._workers:
            for event in self._poll.poll(PROGRESS_INTERVAL):
                worker = self._workers.get(event.fd)
                if worker:
                    if event.can_read:
                        worker.read_logline(event.fd)
                    if event.is_terminated:
                        returncode = worker.terminate()

                        for fd in worker.fds():
                            self._poll.remove(fd)
                            del self._workers[fd]
                        self._num_active_workers -= 1

                        task = worker.task
                        if returncode:
                            # worker did not finish successfully
                            if worker.cleaned or \
                               returncode == METAFLOW_EXIT_DISALLOW_RETRY:
                                self._logger("This failed task will not be "
                                             "retried.", system_msg=True)
                            else:
                                if task.retries < task.user_code_retries +\
                                        task.error_retries:
                                    self._retry_worker(worker)
                                else:
                                    raise TaskFailed(task)
                        else:
                            # worker finished successfully
                            yield task

    def _launch_workers(self):
        while self._run_queue and self._num_active_workers < self._max_workers:
            step, task_kwargs = self._queue_pop()
            # Initialize the task (which can be expensive using remote datastores)
            # before launching the worker so that cost is amortized over time, instead
            # of doing it during _queue_push.
            task = self._new_task(step, **task_kwargs)
            self._launch_worker(task)

    def _retry_worker(self, worker):
        worker.task.retries += 1
        if worker.task.retries >= MAX_ATTEMPTS:
            # any results with an attempt ID >= MAX_ATTEMPTS will be ignored
            # by datastore, so running a task with such a retry_could would
            # be pointless and dangerous
            raise MetaflowInternalError("Too many task attempts (%d)! "
                                        "MAX_ATTEMPTS exceeded."
                                        % worker.task.retries)

        worker.task.new_attempt()
        self._launch_worker(worker.task)

    def _launch_worker(self, task):
        worker = Worker(task, self._max_log_size)
        for fd in worker.fds():
            self._workers[fd] = worker
            self._poll.add(fd)
        self._num_active_workers += 1


class Task(object):
    clone_pathspec_mapping = {}

    def __init__(self,
                 datastore,
                 flow,
                 step,
                 run_id,
                 metadata,
                 environment,
                 entrypoint,
                 event_logger,
                 monitor,
                 input_paths=None,
                 split_index=None,
                 ubf_context=None,
                 clone_run_id=None,
                 origin_ds_set=None,
                 may_clone=False,
                 join_type=None,
                 logger=None,
                 task_id=None,
                 decos=[]):
        if ubf_context == UBF_CONTROL:
            [input_path] = input_paths
            run, input_step, input_task = input_path.split('/')
            # We associate the control task-id to be 1:1 with the split node
            # where the unbounded-foreach was defined.
            # We prefer encoding the corresponding split into the task_id of
            # the control node; so it has access to this information quite
            # easily. There is anyway a corresponding int id stored in the
            # metadata backend - so this should be fine.
            task_id = 'control-%s-%s-%s' % (run, input_step, input_task)
        # Register only regular Metaflow (non control) tasks.
        if task_id is None:
            task_id = str(metadata.new_task_id(run_id, step))
        else:
            # task_id is preset only by persist_parameters() or control tasks.
            if ubf_context == UBF_CONTROL:
                metadata.register_task_id(run_id, step, task_id,
                    sys_tags=[CONTROL_TASK_TAG])
            else:
                metadata.register_task_id(run_id, step, task_id)

        self.step = step
        self.flow_name = flow.name
        self.run_id = run_id
        self.task_id = task_id
        self.input_paths = input_paths
        self.split_index = split_index
        self.ubf_context = ubf_context
        self.decos = decos
        self.entrypoint = entrypoint
        self.environment = environment
        self.environment_type = self.environment.TYPE
        self.clone_run_id = clone_run_id
        self.clone_origin = None
        self.origin_ds_set = origin_ds_set
        self.metadata = metadata
        self.event_logger = event_logger
        self.monitor = monitor

        self._logger = logger
        self._path = '%s/%s/%s' % (self.run_id, self.step, self.task_id)

        self.retries = 0
        self.user_code_retries = 0
        self.error_retries = 0

        self.tags = metadata.sticky_tags
        self.event_logger_type = self.event_logger.logger_type
        self.monitor_type = monitor.monitor_type

        self.metadata_type = metadata.TYPE
        self.datastore_type = datastore.TYPE
        self._datastore = datastore
        self.datastore_sysroot = datastore.datastore_root
        self._results_ds = None

        if clone_run_id and may_clone:
            self._is_cloned = self._attempt_clone(clone_run_id, join_type)
        else:
            self._is_cloned = False

        # Open the output datastore only if the task is not being cloned.
        if not self._is_cloned:
            self.new_attempt()

            for deco in decos:
                deco.runtime_task_created(self._ds,
                                          task_id,
                                          split_index,
                                          input_paths,
                                          self._is_cloned,
                                          ubf_context)

                # determine the number of retries of this task
                user_code_retries, error_retries = deco.step_task_retry_count()
                if user_code_retries is None and error_retries is None:
                    # This signals the runtime that the task doesn't want any
                    # retries indifferent to other decorator opinions.
                    # NOTE: This is needed since we don't statically disallow 
                    # specifying `@retry` in combination with decorators which
                    # implement `unbounded_foreach` semantics. This allows for
                    # ergonomic user invocation of `--with retry`; instead
                    # choosing to specially handle this way in the runtime.
                    self.user_code_retries = None
                    self.error_retries = None
                if self.user_code_retries is not None and \
                    self.error_retries is not None:
                    self.user_code_retries = max(self.user_code_retries,
                                                 user_code_retries)
                    self.error_retries = max(self.error_retries, error_retries)
            if self.user_code_retries is None and self.error_retries is None:
                self.user_code_retries = 0
                self.error_retries = 0

    def new_attempt(self):
        self._ds = self._datastore(self.flow_name,
                                   run_id=self.run_id,
                                   step_name=self.step,
                                   task_id=self.task_id,
                                   mode='w',
                                   metadata=self.metadata,
                                   attempt=self.retries,
                                   event_logger=self.event_logger,
                                   monitor=self.monitor)
        self._ds.init_task()


    def log(self, msg, system_msg=False, pid=None, timestamp=True):
        if pid:
            prefix = '[%s (pid %s)] ' % (self._path, pid)
        else:
            prefix = '[%s] ' % self._path

        self._logger(msg,
                     head=prefix,
                     system_msg=system_msg,
                     timestamp=timestamp)
        sys.stdout.flush()

    def _find_origin_task(self, clone_run_id, join_type):
        if self.step == '_parameters':
            pathspec = '%s/_parameters[]' % clone_run_id
            origin = self.origin_ds_set.get_with_pathspec_index(pathspec)

            if origin is None:
                # This is just for usability: We could rerun the whole flow
                # if an unknown clone_run_id is provided but probably this is
                # not what the user intended, so raise a warning
                raise MetaflowException("Resume could not find run id *%s*" %
                                        clone_run_id)
            else:
                return origin
        else:
            # all inputs must have the same foreach stack, so we can safely
            # pick the first one
            parent_pathspec = self.input_paths[0]
            origin_parent_pathspec = \
                self.clone_pathspec_mapping[parent_pathspec]
            parent = self.origin_ds_set.get_with_pathspec(origin_parent_pathspec)
            # Parent should be non-None since only clone the child if the parent
            # was successfully cloned.
            foreach_stack = parent['_foreach_stack']
            if join_type == 'foreach':
                # foreach-join pops the topmost index
                index = ','.join(str(s.index) for s in foreach_stack[:-1])
            elif self.split_index or self.ubf_context == UBF_CONTROL:
                # foreach-split pushes a new index
                index = ','.join([str(s.index) for s in foreach_stack] +
                                 [str(self.split_index)])
            else:
                # all other transitions keep the parent's foreach stack intact
                index = ','.join(str(s.index) for s in foreach_stack)
            pathspec = '%s/%s[%s]' % (clone_run_id, self.step, index)
            return self.origin_ds_set.get_with_pathspec_index(pathspec)

    def _attempt_clone(self, clone_run_id, join_type):
        origin = self._find_origin_task(clone_run_id, join_type)

        if origin and origin['_task_ok']:
            # Store the mapping from current_pathspec -> origin_pathspec which
            # will be useful for looking up origin_ds_set in find_origin_task.
            self.clone_pathspec_mapping[self._path] = origin.pathspec
            if self.step == '_parameters':
                # Clone in place without relying on run_queue.
                self.new_attempt()
                self._ds.clone(origin)
                self._ds.done()
            else:
                self.log("Cloning results of a previously run task %s"
                         % origin.pathspec, system_msg=True)
                # Store the origin pathspec in clone_origin so this can be run
                # as a task by the runtime.
                self.clone_origin = origin.pathspec
                # Save a call to creating the results_ds since its same as origin.
                self._results_ds = origin
            return True
        else:
            return False

    @property
    def path(self):
        return self._path

    @property
    def results(self):
        if self._results_ds:
            return self._results_ds
        else:
            self._results_ds = self._datastore(self.flow_name,
                                               run_id=self.run_id,
                                               step_name=self.step,
                                               task_id=self.task_id,
                                               mode='r',
                                               metadata=self.metadata,
                                               event_logger=self.event_logger,
                                               monitor=self.monitor)
            return self._results_ds

    @property
    def finished_id(self):
        # note: id is not available before the task has finished
        return (self.step, tuple(self.results['_foreach_stack']))

    @property
    def is_cloned(self):
        return self._is_cloned

    def persist(self, flow):
        # this is used to persist parameters before the start step
        flow._task_ok = flow._success = True
        flow._foreach_stack = []
        self._ds.persist(flow)
        self._ds.done()

    def save_logs(self, stdout_buffer, stderr_buffer):
        self._ds.save_logs(RUNTIME_LOG_SOURCE, [
            ('stdout', stdout_buffer),
            ('stderr', stderr_buffer)
        ])

    def save_metadata(self, name, metadata):
        self._ds.save_metadata(name, metadata)

    def __str__(self):
        return ' '.join(self._args)


class TaskFailed(MetaflowException):
    headline = "Step failure"

    def __init__(self, task, msg=''):
        body = "Step *%s* (task-id %s) failed" % (task.step,
                                                  task.task_id)
        if msg:
            body = '%s: %s' % (body, msg)
        else:
            body += '.'

        super(TaskFailed, self).__init__(body)


class TruncatedBuffer(object):
    def __init__(self, name, maxsize):
        self.name = name
        self._maxsize = maxsize
        self._buffer = BytesIO()
        self._size = 0
        self._eof = False

    def write(self, bytedata, system_msg=False):
        if system_msg:
            self._buffer.write(bytedata)
        elif not self._eof:
            if self._size + len(bytedata) < self._maxsize:
                self._buffer.write(bytedata)
                self._size += len(bytedata)
            else:
                msg = b'[TRUNCATED - MAXIMUM LOG FILE SIZE REACHED]\n'
                self._buffer.write(mflog_msg(msg))
                self._eof = True

    def get_bytes(self):
        return self._buffer.getvalue()


class CLIArgs(object):
    """
    Container to allow decorators modify the command line parameters
    for step execution in StepDecorator.runtime_step_cli().
    """

    def __init__(self, task):
        self.task = task
        self.entrypoint = list(task.entrypoint)
        self.top_level_options = {
            'quiet': True,
            'coverage': 'coverage' in sys.modules,
            'metadata': self.task.metadata_type,
            'environment': self.task.environment_type,
            'datastore': self.task.datastore_type,
            'event-logger': self.task.event_logger_type,
            'monitor': self.task.monitor_type,
            'datastore-root': self.task.datastore_sysroot,
            'with': [deco.make_decorator_spec() for deco in self.task.decos
                     if not deco.statically_defined]
        }

        # FlowDecorators can define their own top-level options. They are
        # responsible for adding their own top-level options and values through
        # the get_top_level_options() hook.
        for deco in flow_decorators():
            self.top_level_options.update(deco.get_top_level_options())

        self.commands = ['step']
        self.command_args = [self.task.step]
        self.command_options = {
            'run-id': task.run_id,
            'task-id': task.task_id,
            'input-paths': compress_list(task.input_paths),
            'split-index': task.split_index,
            'retry-count': task.retries,
            'max-user-code-retries': task.user_code_retries,
            'tag': task.tags,
            'namespace': get_namespace() or '',
            'ubf-context': task.ubf_context
        }
        self.env = {}

    def get_args(self):

        # TODO: Make one with dict_to_cli_options; see cli_args.py for more detail
        def _options(mapping):
            for k, v in mapping.items():
                if v:
                    # we need special handling for 'with' since it is a reserved
                    # keyword in Python, so we call it 'decospecs' in click args
                    if k == 'decospecs':
                        k = 'with'
                    k = k.replace('_', '-')
                    v = v if isinstance(v, (list, tuple, set)) else [v]
                    for value in v:
                        yield '--%s' % k
                        if not isinstance(value, bool):
                            yield to_unicode(value)

        args = list(self.entrypoint)
        args.extend(_options(self.top_level_options))
        args.extend(self.commands)
        args.extend(self.command_args)
        args.extend(_options(self.command_options))
        return args

    def get_env(self):
        return self.env

    def __str__(self):
        return ' '.join(self.get_args())


class Worker(object):

    def __init__(self, task, max_logs_size):

        self.task = task
        self._proc = self._launch()

        if task.retries > task.user_code_retries:
            self.task.log('Task fallback is starting to handle the failure.',
                          system_msg=True,
                          pid=self._proc.pid)
        elif not task.is_cloned:
            suffix = ' (retry).' if task.retries else '.'
            self.task.log('Task is starting' + suffix,
                          system_msg=True,
                          pid=self._proc.pid)

        self._stdout = TruncatedBuffer('stdout', max_logs_size)
        self._stderr = TruncatedBuffer('stderr', max_logs_size)

        self._logs = {self._proc.stderr.fileno(): (self._proc.stderr,
                                                   self._stderr),
                      self._proc.stdout.fileno(): (self._proc.stdout,
                                                   self._stdout)}

        self._encoding = sys.stdout.encoding or 'UTF-8'
        self.killed = False  # Killed indicates that the task was forcibly killed
                             # with SIGKILL by the master process.
                             # A killed task is always considered cleaned
        self.cleaned = False  # A cleaned task is one that is shutting down and has been
                              # noticed by the runtime and queried for its state (whether or
                              # not is is properly shut down)

    def _launch(self):
        args = CLIArgs(self.task)
        env = dict(os.environ)

        if self.task.clone_run_id:
            args.command_options['clone-run-id'] = self.task.clone_run_id

        if self.task.is_cloned and self.task.clone_origin:
            args.command_options['clone-only'] = self.task.clone_origin
            # disabling atlas sidecar for cloned tasks due to perf reasons
            args.top_level_options['monitor'] = 'nullSidecarMonitor'
        else:
            # decorators may modify the CLIArgs object in-place
            for deco in self.task.decos:
                deco.runtime_step_cli(args,
                                      self.task.retries,
                                      self.task.user_code_retries,
                                      self.task.ubf_context)
        env.update(args.get_env())
        env['PYTHONUNBUFFERED'] = 'x'
        # NOTE bufsize=1 below enables line buffering which is required
        # by read_logline() below that relies on readline() not blocking
        # print('running', args)
        cmdline = args.get_args()
        debug.subcommand_exec(cmdline)
        return subprocess.Popen(cmdline,
                                env=env,
                                bufsize=1,
                                stdin=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                stdout=subprocess.PIPE)

    def emit_log(self, msg, buf, system_msg=False):
        if mflog.is_structured(msg):
            res = mflog.parse(msg)
            if res:
                # parsing successful
                plain = res.msg
                timestamp = res.utc_tstamp
                if res.should_persist:
                    # in special circumstances we may receive structured
                    # loglines that haven't been properly persisted upstream.
                    # This is the case e.g. if a task crashes in an external
                    # system and we retrieve the remaining logs after the crash.
                    # Those lines are marked with a special tag, should_persist,
                    # which we process here
                    buf.write(mflog.unset_should_persist(msg))
            else:
                # parsing failed, corrupted logline. Print it as-is
                timestamp = datetime.utcnow()
                plain = msg
        else:
            # If the line isn't formatted with mflog already, we format it here.
            plain = msg
            timestamp = datetime.utcnow()
            # store unformatted loglines in the buffer that will be
            # persisted, assuming that all previously formatted loglines have
            # been already persisted at the source.
            buf.write(mflog_msg(msg, now=timestamp), system_msg=system_msg)
        text = plain.strip().decode(self._encoding, errors='replace')
        self.task.log(text,
                      pid=self._proc.pid,
                      timestamp=mflog.utc_to_local(timestamp),
                      system_msg=system_msg)

    def read_logline(self, fd):
        fileobj, buf = self._logs[fd]
        # readline() below should never block thanks to polling and
        # line buffering. If it does, things will deadlock
        line = fileobj.readline()
        if line:
            self.emit_log(line, buf)
            return True
        else:
            return False

    def fds(self):
        return (self._proc.stderr.fileno(),
                self._proc.stdout.fileno())

    def clean(self):
        if self.killed:
            return True
        if not self.cleaned:
            for fileobj, buf in self._logs.values():
                msg = b'[KILLED BY ORCHESTRATOR]\n'
                self.emit_log(msg, buf, system_msg=True)
            self.cleaned = True
        return self._proc.poll() is not None

    def kill(self):
        if not self.killed:
            try:
                self._proc.kill()
            except:
                pass
            self.cleaned = True
            self.killed = True

    def terminate(self):
        # this shouldn't block, since terminate() is called only
        # after the poller has decided that the worker is dead
        returncode = self._proc.wait()

        # consume all remaining loglines
        # we set the file descriptor to be non-blocking, since
        # the pipe may stay active due to subprocesses launched by
        # the worker, e.g. sidecars, so we can't rely on EOF. We try to
        # read just what's available in the pipe buffer
        for fileobj, buf in self._logs.values():
            fileno = fileobj.fileno()
            fcntl.fcntl(fileno, fcntl.F_SETFL, os.O_NONBLOCK)
            try:
                while self.read_logline(fileno):
                    pass
            except:
                # ignore "resource temporarily unavailable" etc. errors
                # caused due to non-blocking. Draining is done on a
                # best-effort basis.
                pass

        # Return early if the task is cloned since we don't want to
        # perform any log collection.
        if not self.task.is_cloned:
            self.task.save_metadata('runtime', {'return_code': returncode,
                                                'killed': self.killed,
                                                'success': returncode == 0})
            if returncode:
                if not self.killed:
                    if returncode == -11:
                        self.emit_log(b'Task failed with a segmentation fault.',
                                      self._stderr,
                                      system_msg=True)
                    else:
                        self.emit_log(b'Task failed.',
                                      self._stderr,
                                      system_msg=True)
            else:
                num = self.task.results['_foreach_num_splits']
                if num:
                    self.task.log('Foreach yields %d child steps.' % num,
                                  system_msg=True,
                                  pid=self._proc.pid)
                self.task.log('Task finished successfully.',
                              system_msg=True,
                              pid=self._proc.pid)
            self.task.save_logs(self._stdout.get_bytes(),
                                self._stderr.get_bytes())
        return returncode

    def __str__(self):
        return 'Worker[%d]: %s' % (self._proc.pid, self.task.path)



import json
from collections import namedtuple

import click

from .util import get_username, is_stringish
from .exception import ParameterFieldFailed,\
                       ParameterFieldTypeMismatch,\
                       MetaflowException

try:
    # Python2
    strtype = basestring
except:
    # Python3
    strtype = str

# ParameterContext allows deploy-time functions modify their
# behavior based on the context. We can add fields here without
# breaking backwards compatibility but don't remove any fields!
ParameterContext = namedtuple('ParameterContext',
                              ['flow_name',
                               'user_name',
                               'parameter_name',
                               'logger',
                               'ds_type'])

# currently we execute only one flow per process, so we can treat
# Parameters globally. If this was to change, it should/might be
# possible to move these globals in a FlowSpec (instance) specific
# closure.
parameters = []
context_proto = None

class JSONTypeClass(click.ParamType):
    name = 'JSON'

    def convert(self, value, param, ctx):
        if not isinstance(value, strtype):
            # Already a correct type
            return value
        try:
            return json.loads(value)
        except:
            self.fail("%s is not a valid JSON object" % value, param, ctx)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return 'JSON'

class DeployTimeField(object):
    """
    This a wrapper object for a user-defined function that is called
    at the deploy time to populate fields in a Parameter. The wrapper
    is needed to make Click show the actual value returned by the
    function instead of a function pointer in its help text. Also this
    object curries the context argument for the function, and pretty
    prints any exceptions that occur during evaluation.
    """
    def __init__(self,
                 parameter_name,
                 parameter_type,
                 field,
                 fun,
                 return_str=True,
                 print_representation=None):

        self.fun = fun
        self.field = field
        self.parameter_name = parameter_name
        self.parameter_type = parameter_type
        self.return_str = return_str
        self.print_representation = self.user_print_representation = print_representation
        if self.print_representation is None:
            self.print_representation = str(self.fun)

    def __call__(self, full_evaluation=False):
        # full_evaluation is True if there will be no further "convert" called
        # by click and the parameter should be fully evaluated.
        ctx = context_proto._replace(parameter_name=self.parameter_name)
        try:
            try:
                # Not all functions take two arguments
                val = self.fun(ctx, full_evaluation)
            except TypeError:
                val = self.fun(ctx)
        except:
            raise ParameterFieldFailed(self.parameter_name, self.field)
        else:
            return self._check_type(val)

    def _check_type(self, val):
        # it is easy to introduce a deploy-time function that that accidentally
        # returns a value whose type is not compatible with what is defined
        # in Parameter. Let's catch those mistakes early here, instead of
        # showing a cryptic stack trace later.

        # note: this doesn't work with long in Python2 or types defined as
        # click types, e.g. click.INT
        TYPES = {bool: 'bool',
                 int: 'int',
                 float: 'float',
                 list: 'list'}

        msg = "The value returned by the deploy-time function for "\
              "the parameter *%s* field *%s* has a wrong type. " %\
              (self.parameter_name, self.field)

        if self.parameter_type in TYPES:
            if type(val) != self.parameter_type:
                msg += 'Expected a %s.' % TYPES[self.parameter_type]
                raise ParameterFieldTypeMismatch(msg)
            return str(val) if self.return_str else val
        else:
            if not is_stringish(val):
                msg += 'Expected a string.'
                raise ParameterFieldTypeMismatch(msg)
            return val

    @property
    def description(self):
        return self.print_representation

    def __str__(self):
        if self.user_print_representation:
            return self.user_print_representation
        return self()

    def __repr__(self):
        if self.user_print_representation:
            return self.user_print_representation
        return self()


def deploy_time_eval(value):
    if isinstance(value, DeployTimeField):
        return value(full_evaluation=True)
    else:
        return value

# this is called by cli.main
def set_parameter_context(flow_name, echo, datastore):
    global context_proto
    context_proto = ParameterContext(flow_name=flow_name,
                                     user_name=get_username(),
                                     parameter_name=None,
                                     logger=echo,
                                     ds_type=datastore.TYPE)

class Parameter(object):
    def __init__(self, name, **kwargs):
        self.name = name
        self.kwargs = kwargs
        # TODO: check that the type is one of the supported types
        param_type = self.kwargs['type'] = self._get_type(kwargs)

        if self.name == 'params':
            raise MetaflowException("Parameter name 'params' is a reserved "
                                    "word. Please use a different "
                                    "name for your parameter.")

        # make sure the user is not trying to pass a function in one of the
        # fields that don't support function-values yet
        for field in ('show_default',
                      'separator',
                      'required'):
            if callable(kwargs.get(field)):
                raise MetaflowException("Parameter *%s*: Field '%s' cannot "
                                        "have a function as its value"\
                                        % (name, field))

        self.kwargs['show_default'] = self.kwargs.get('show_default', True)

        # default can be defined as a function
        default_field = self.kwargs.get('default')
        if callable(default_field) and not isinstance(default_field, DeployTimeField):
            self.kwargs['default'] = DeployTimeField(name,
                                                     param_type,
                                                     'default',
                                                     self.kwargs['default'],
                                                     return_str=True)

        # note that separator doesn't work with DeployTimeFields unless you
        # specify type=str
        self.separator = self.kwargs.pop('separator', None)
        if self.separator and not self.is_string_type:
            raise MetaflowException("Parameter *%s*: Separator is only allowed "
                                    "for string parameters." % name)
        parameters.append(self)

    def option_kwargs(self, deploy_mode):
        kwargs = self.kwargs
        if isinstance(kwargs.get('default'), DeployTimeField) and not deploy_mode:
            ret = dict(kwargs)
            help_msg = kwargs.get('help')
            help_msg = '' if help_msg is None else help_msg
            ret['help'] = help_msg + \
                "[default: deploy-time value of '%s']" % self.name
            ret['default'] = None
            ret['required'] = False
            return ret
        else:
            return kwargs

    def load_parameter(self, v):
        return v

    def _get_type(self, kwargs):
        default_type = str

        default = kwargs.get('default')
        if default is not None and not callable(default):
            default_type = type(default)

        return kwargs.get('type', default_type)

    @property
    def is_string_type(self):
        return self.kwargs.get('type', str) == str and\
               isinstance(self.kwargs.get('default', ''), strtype)

    # this is needed to appease Pylint for JSONType'd parameters,
    # which may do self.param['foobar']
    def __getitem__(self, x):
        pass

def add_custom_parameters(deploy_mode=False):
    # deploy_mode determines whether deploy-time functions should or should
    # not be evaluated for this command
    def wrapper(cmd):
        # Iterate over parameters in reverse order so cmd.params lists options
        # in the order they are defined in the FlowSpec subclass
        for arg in parameters[::-1]:
            kwargs = arg.option_kwargs(deploy_mode)
            cmd.params.insert(0, click.Option(('--' + arg.name,), **kwargs))
        return cmd
    return wrapper

def set_parameters(flow, kwargs):
    seen = set()
    for var, param in flow._get_parameters():
        norm = param.name.lower()
        if norm in seen:
            raise MetaflowException("Parameter *%s* is specified twice. "
                                    "Note that parameter names are "
                                    "case-insensitive." % param.name)
        seen.add(norm)

    flow._success = True
    for var, param in flow._get_parameters():
        val = kwargs[param.name.replace('-', '_').lower()]
        # Support for delayed evaluation of parameters. This is used for
        # includefile in particular
        if callable(val):
            val = val()
        val = val.split(param.separator) if val and param.separator else val
        setattr(flow, var, val)



from __future__ import print_function

import sys

from metaflow.sidecar_messages import MessageTypes, Message
from metaflow.monitor import Timer, deserialize_metric
from metaflow.monitor import MEASURE_TYPE, get_monitor_msg_type

class DebugMonitor(object):
    TYPE = 'debugMonitor'

    def __init__(self):
        self.logger('init')

    def count(self, count):
        pass

    def measure(self, timer):
        # type: (Timer) -> None
        self.logger('elapsed time for {}: {}'.
                    format(timer.name, str(timer.get_duration())))

    def gauge(self, gauge):
        pass

    def process_message(self, msg):
        # type: (Message) -> None
        self.logger('processing message %s' % str(msg.msg_type))
        msg_type = get_monitor_msg_type(msg)
        if msg_type == MEASURE_TYPE:
            timer = deserialize_metric(msg.payload.get('timer'))
            self.measure(timer)
        else:
            pass

    def shutdown(self):
        sys.stderr.flush()

    def logger(self, msg):
        print('local_monitor: %s' % msg, file=sys.stderr)
        sys.stderr.flush()



import errno
import os
import json
import subprocess
import time
from distutils.version import LooseVersion

from metaflow.exception import MetaflowException
from metaflow.metaflow_environment import InvalidEnvironmentException
from metaflow.util import which

class CondaException(MetaflowException):
    headline = 'Conda ran into an error while setting up environment.'

    def __init__(self, error):
        if isinstance(error, (list,)):
            error = '\n'.join(error)
        msg = '{error}'.format(error=error)
        super(CondaException, self).__init__(msg)

class CondaStepException(CondaException):

    def __init__(self, exception, step):
        msg = 'Step: {step}, Error: {error}'.format(step=step, error=exception.message)
        super(CondaStepException, self).__init__(msg)

class Conda(object):

    def __init__(self):
        self._bin = which('conda')
        if self._bin is None:
            raise InvalidEnvironmentException('No conda installation found. '
                                              'Install conda first.')
        if LooseVersion(self._info()['conda_version']) < LooseVersion('4.6.0'):
            raise InvalidEnvironmentException('Conda version 4.6.0 or newer '
                                              'is required. Visit '
                                              'https://docs.conda.io/en/latest/miniconda.html '
                                              'for installation instructions.')
        if 'conda-forge' not in self.config()['channels']:
            raise InvalidEnvironmentException('Conda channel \'conda-forge\' '
                                              'is required. Specify it with CONDA_CHANNELS '
                                              'environment variable.')
    
    def create(self,
               step_name,
               env_id,
               deps,
               architecture=None,
               explicit=False,
               disable_safety_checks=False):
        # Create the conda environment
        try:
            with CondaLock(self._env_lock_file(env_id)):
                self._remove(env_id)
                self._create(env_id,
                             deps,
                             explicit,
                             architecture,
                             disable_safety_checks)
                return self._deps(env_id)
        except CondaException as e:
            raise CondaStepException(e, step_name)

    def remove(self, step_name, env_id):
        # Remove the conda environment
        try:
            with CondaLock(self._env_lock_file(env_id)):
                self._remove(env_id)
        except CondaException as e:
            raise CondaStepException(e, step_name)

    def python(self, env_id):
        # Get Python interpreter for the conda environment
        return os.path.join(self._env_path(env_id), 'bin/python')

    def environments(self, flow):
        # List all conda environments associated with the flow
        envs = self._info()['envs']
        ret = {}
        for env in envs:
            if '/envs/' in env:
                name = os.path.basename(env)
                if name.startswith('metaflow_%s' % flow):
                    ret[name] = env
        return ret

    def config(self):
        # Show conda installation configuration
        return json.loads(self._call_conda(['config', '--show']))

    def package_info(self, env_id):
        # Show conda environment package configuration
        # Not every parameter is exposed via conda cli hence this ignominy
        metadata = os.path.join(self._env_path(env_id), 'conda-meta')
        for path, dirs, files in os.walk(metadata):
            for file in files:
                if file.endswith('.json'):
                    with open(os.path.join(path, file)) as f:
                        yield json.loads(f.read())

    def _info(self):
        return json.loads(self._call_conda(['info']))

    def _create(self,
                env_id,
                deps,
                explicit=False,
                architecture=None,
                disable_safety_checks=False):
        cmd = ['create', '--yes', '--no-default-packages',
               '--name', env_id, '--quiet']
        if explicit:
            cmd.append('--no-deps')
        cmd.extend(deps)
        self._call_conda(cmd,
                         architecture=architecture,
                         disable_safety_checks=disable_safety_checks)

    def _remove(self, env_id):
        self._call_conda(['env', 'remove', '--name',
                           env_id, '--yes', '--quiet'])

    def _install(self, env_id, deps, explicit=False):
        cmd = ['install', '--yes', '--name', env_id, '--quiet']
        if explicit:
            cmd.append('--no-deps')
        cmd.extend(deps)
        self._call_conda(cmd)

    def _install_order(self, env_id):
        cmd = ['list', '--name', env_id, '--explicit']
        response = self._call_conda(cmd).decode('utf-8') 
        emit = False
        result = []
        for line in response.splitlines():
            if emit:
                result.append(line.split('/')[-1])
            if not emit and line == '@EXPLICIT':
                emit = True  
        return result

    def _deps(self, env_id):
        exact_deps = []
        urls = []
        for package in self.package_info(env_id):
            exact_deps.append('%s=%s=%s' % (package['name'], package['version'], package['build']))
            urls.append(package['url'])
        order = self._install_order(env_id)
        return (exact_deps, urls, order)

    def _env_path(self, env_id):
        envs = self._info()['envs']
        for env in envs:
            if '/envs/' in env:
                name = os.path.basename(env)
                if name == env_id:
                    return env
        return None

    def _env_lock_file(self, env_id):
        return os.path.join(self._info()['envs_dirs'][0], 'mf_env-creation.lock')

    def _call_conda(self, args, architecture=None, disable_safety_checks=False):
        try:
            env = {
                'CONDA_JSON': 'True',
                'CONDA_SUBDIR': (architecture if architecture else ''),
                'CONDA_USE_ONLY_TAR_BZ2': 'True'
            }
            if disable_safety_checks:
                env['CONDA_SAFETY_CHECKS'] = 'disabled'
            return subprocess.check_output(
                [self._bin] + args, 
                stderr = open(os.devnull, 'wb'),
                env = dict(os.environ, **env)).strip()
        except subprocess.CalledProcessError as e:
            try:
                output = json.loads(e.output)
                err = [output['error']]
                for error in output.get('errors', []):
                    err.append(error['error'])
                raise CondaException(err)
            except (TypeError, ValueError) as ve:
                pass
            raise CondaException(
                'command \'{cmd}\' returned error ({code}): {output}'
                    .format(cmd=e.cmd, code=e.returncode, output=e.output))


class CondaLock(object):

    def __init__(self, lock, timeout=3600, delay=10):
        self.lock = lock
        self.locked = False
        self.timeout = timeout
        self.delay = delay

    def _acquire(self):
        start = time.time()
        try:
            os.makedirs(os.path.dirname(self.lock))
        except OSError as x:
            if x.errno != errno.EEXIST:
                raise
        while True:
            try:
                self.fd = os.open(self.lock, os.O_CREAT |
                                  os.O_EXCL | os.O_RDWR)
                self.locked = True
                break
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
                if self.timeout is None:
                    raise CondaException(
                        'Could not acquire lock {}'.format(self.lock))
                if (time.time() - start) >= self.timeout:
                    raise CondaException(
                        'Timeout occured while acquiring lock {}'.format(self.lock))
                time.sleep(self.delay)

    def _release(self):
        if self.locked:
            os.close(self.fd)
            os.unlink(self.lock)
            self.locked = False

    def __enter__(self):
        if not self.locked:
            self._acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.__del__()

    def __del__(self):
        self._release()


import os
import time
import json
import select
import atexit
import shlex
import time
import warnings

from requests.exceptions import HTTPError
from metaflow.exception import MetaflowException, MetaflowInternalError
from metaflow.metaflow_config import BATCH_METADATA_SERVICE_URL, DATATOOLS_S3ROOT, \
    DATASTORE_LOCAL_DIR, DATASTORE_SYSROOT_S3, DEFAULT_METADATA, \
    BATCH_METADATA_SERVICE_HEADERS, BATCH_EMIT_TAGS
from metaflow import util

from .batch_client import BatchClient

from metaflow.datastore.util.s3tail import S3Tail
from metaflow.mflog.mflog import refine, set_should_persist
from metaflow.mflog import export_mflog_env_vars,\
                           bash_capture_logs,\
                           update_delay,\
                           BASH_SAVE_LOGS

# Redirect structured logs to /logs/
LOGS_DIR = '/logs'
STDOUT_FILE = 'mflog_stdout'
STDERR_FILE = 'mflog_stderr'
STDOUT_PATH = os.path.join(LOGS_DIR, STDOUT_FILE)
STDERR_PATH = os.path.join(LOGS_DIR, STDERR_FILE)

class BatchException(MetaflowException):
    headline = 'AWS Batch error'


class BatchKilledException(MetaflowException):
    headline = 'AWS Batch task killed'


class Batch(object):
    def __init__(self, metadata, environment):
        self.metadata = metadata
        self.environment = environment
        self._client = BatchClient()
        atexit.register(lambda: self.job.kill() if hasattr(self, 'job') else None)

    def _command(self,
                 environment,
                 code_package_url,
                 step_name,
                 step_cmds,
                 task_spec):
        mflog_expr = export_mflog_env_vars(datastore_type='s3',
                                           stdout_path=STDOUT_PATH,
                                           stderr_path=STDERR_PATH,
                                           **task_spec)
        init_cmds = environment.get_package_commands(code_package_url)
        init_expr = ' && '.join(init_cmds)
        step_expr = bash_capture_logs(' && '.join(
                        environment.bootstrap_commands(step_name) + step_cmds))

        # construct an entry point that
        # 1) initializes the mflog environment (mflog_expr)
        # 2) bootstraps a metaflow environment (init_expr)
        # 3) executes a task (step_expr)

        # the `true` command is to make sure that the generated command
        # plays well with docker containers which have entrypoint set as
        # eval $@
        cmd_str = 'true && mkdir -p /logs && %s && %s && %s; ' % \
                        (mflog_expr, init_expr, step_expr)
        # after the task has finished, we save its exit code (fail/success)
        # and persist the final logs. The whole entrypoint should exit
        # with the exit code (c) of the task.
        #
        # Note that if step_expr OOMs, this tail expression is never executed.
        # We lose the last logs in this scenario (although they are visible 
        # still through AWS CloudWatch console).
        cmd_str += 'c=$?; %s; exit $c' % BASH_SAVE_LOGS
        return shlex.split('bash -c \"%s\"' % cmd_str)

    def _search_jobs(self, flow_name, run_id, user):
        if user is None:
            regex = '-{flow_name}-'.format(flow_name=flow_name)
        else:
            regex = '{user}-{flow_name}-'.format(
                user=user, flow_name=flow_name
            )
        jobs = []
        for job in self._client.unfinished_jobs():
            if regex in job['jobName']:
                jobs.append(job['jobId'])
        if run_id is not None:
            run_id = run_id[run_id.startswith('sfn-') and len('sfn-'):]
        for job in self._client.describe_jobs(jobs):
            parameters = job['parameters']
            match = (user is None or parameters['metaflow.user'] == user) and \
                    (parameters['metaflow.flow_name'] == flow_name) and \
                    (run_id is None or parameters['metaflow.run_id'] == run_id)
            if match:
                yield job

    def _job_name(self, user, flow_name, run_id, step_name, task_id, retry_count):
        return '{user}-{flow_name}-{run_id}-{step_name}-{task_id}-{retry_count}'.format(
            user=user,
            flow_name=flow_name,
            run_id=str(run_id) if run_id is not None else '',
            step_name=step_name,
            task_id=str(task_id) if task_id is not None else '',
            retry_count=str(retry_count) if retry_count is not None else ''
        )

    def list_jobs(self, flow_name, run_id, user, echo):
        jobs = self._search_jobs(flow_name, run_id, user)
        found = False
        for job in jobs:
            found = True
            echo(
                '{name} [{id}] ({status})'.format(
                    name=job['jobName'], id=job['jobId'], status=job['status']
                )
            )
        if not found:
            echo('No running AWS Batch jobs found.')

    def kill_jobs(self, flow_name, run_id, user, echo):
        jobs = self._search_jobs(flow_name, run_id, user)
        found = False
        for job in jobs:
            found = True
            try:
                self._client.attach_job(job['jobId']).kill()
                echo(
                    'Killing AWS Batch job: {name} [{id}] ({status})'.format(
                        name=job['jobName'], id=job['jobId'], status=job['status']
                    )
                )
            except Exception as e:
                echo(
                    'Failed to terminate AWS Batch job %s [%s]'
                    % (job['jobId'], repr(e))
                )
        if not found:
            echo('No running AWS Batch jobs found.')

    def create_job(
        self,
        step_name,
        step_cli,
        task_spec,
        code_package_sha,
        code_package_url,
        code_package_ds,
        image,
        queue,
        iam_role=None,
        execution_role=None,
        cpu=None,
        gpu=None,
        memory=None,
        run_time_limit=None,
        shared_memory=None,
        max_swap=None,
        swappiness=None,
        env={},
        attrs={}
    ):
        job_name = self._job_name(
            attrs.get('metaflow.user'),
            attrs.get('metaflow.flow_name'),
            attrs.get('metaflow.run_id'),
            attrs.get('metaflow.step_name'),
            attrs.get('metaflow.task_id'),
            attrs.get('metaflow.retry_count')
        )
        job = self._client.job()
        job \
            .job_name(job_name) \
            .job_queue(queue) \
            .command(
                self._command(self.environment, code_package_url,
                              step_name, [step_cli], task_spec)) \
            .image(image) \
            .iam_role(iam_role) \
            .execution_role(execution_role) \
            .job_def(image, iam_role,
                queue, execution_role, shared_memory,
                max_swap, swappiness) \
            .cpu(cpu) \
            .gpu(gpu) \
            .memory(memory) \
            .shared_memory(shared_memory) \
            .max_swap(max_swap) \
            .swappiness(swappiness) \
            .timeout_in_secs(run_time_limit) \
            .environment_variable('AWS_DEFAULT_REGION', self._client.region()) \
            .environment_variable('METAFLOW_CODE_SHA', code_package_sha) \
            .environment_variable('METAFLOW_CODE_URL', code_package_url) \
            .environment_variable('METAFLOW_CODE_DS', code_package_ds) \
            .environment_variable('METAFLOW_USER', attrs['metaflow.user']) \
            .environment_variable('METAFLOW_SERVICE_URL', BATCH_METADATA_SERVICE_URL) \
            .environment_variable('METAFLOW_SERVICE_HEADERS', json.dumps(BATCH_METADATA_SERVICE_HEADERS)) \
            .environment_variable('METAFLOW_DATASTORE_SYSROOT_S3', DATASTORE_SYSROOT_S3) \
            .environment_variable('METAFLOW_DATATOOLS_S3ROOT', DATATOOLS_S3ROOT) \
            .environment_variable('METAFLOW_DEFAULT_DATASTORE', 's3') \
            .environment_variable('METAFLOW_DEFAULT_METADATA', DEFAULT_METADATA)
            # Skip setting METAFLOW_DATASTORE_SYSROOT_LOCAL because metadata sync between the local user 
            # instance and the remote AWS Batch instance assumes metadata is stored in DATASTORE_LOCAL_DIR 
            # on the remote AWS Batch instance; this happens when METAFLOW_DATASTORE_SYSROOT_LOCAL 
            # is NOT set (see get_datastore_root_from_config in datastore/local.py).
        for name, value in env.items():
            job.environment_variable(name, value)
        if attrs:
            for key, value in attrs.items():
                job.parameter(key, value)
        # Tags for AWS Batch job (for say cost attribution)
        if BATCH_EMIT_TAGS:
            for key in ['metaflow.flow_name', 'metaflow.run_id',
                            'metaflow.step_name', 'metaflow.version', 
                            'metaflow.run_id.$', 'metaflow.user',
                            'metaflow.owner', 'metaflow.production_token']:
                if key in attrs:
                    job.tag(key, attrs.get(key))
        return job

    def launch_job(
        self,
        step_name,
        step_cli,
        task_spec,
        code_package_sha,
        code_package_url,
        code_package_ds,
        image,
        queue,
        iam_role=None,
        execution_role=None, # for FARGATE compatibility
        cpu=None,
        gpu=None,
        memory=None,
        platform=None,
        run_time_limit=None,
        shared_memory=None,
        max_swap=None,
        swappiness=None,
        env={},
        attrs={},
        ):
        if queue is None:
            queue = next(self._client.active_job_queues(), None)
            if queue is None:
                raise BatchException(
                    'Unable to launch AWS Batch job. No job queue '
                    ' specified and no valid & enabled queue found.'
                )
        job = self.create_job(
                        step_name,
                        step_cli,
                        task_spec,
                        code_package_sha,
                        code_package_url,
                        code_package_ds,
                        image,
                        queue,
                        iam_role,
                        execution_role,
                        cpu,
                        gpu,
                        memory,
                        run_time_limit,
                        shared_memory,
                        max_swap,
                        swappiness,
                        env,
                        attrs
        )
        self.job = job.execute()

    def wait(self, stdout_location, stderr_location, echo=None):
        
        def wait_for_launch(job):
            status = job.status
            echo('Task is starting (status %s)...' % status,
                 'stderr',
                 batch_id=job.id)
            t = time.time()
            while True:
                if status != job.status or (time.time()-t) > 30:
                    status = job.status
                    echo(
                        'Task is starting (status %s)...' % status,
                        'stderr',
                        batch_id=job.id
                    )
                    t = time.time()
                if job.is_running or job.is_done or job.is_crashed:
                    break
                select.poll().poll(200)

        prefix = b'[%s] ' % util.to_bytes(self.job.id)
        
        def _print_available(tail, stream, should_persist=False):
            # print the latest batch of lines from S3Tail
            try:
                for line in tail:
                    if should_persist:
                        line = set_should_persist(line)
                    else:
                        line = refine(line, prefix=prefix)
                    echo(line.strip().decode('utf-8', errors='replace'), stream)
            except Exception as ex:
                echo('[ temporary error in fetching logs: %s ]' % ex,
                     'stderr',
                     batch_id=self.job.id)
        stdout_tail = S3Tail(stdout_location)
        stderr_tail = S3Tail(stderr_location)

        # 1) Loop until the job has started
        wait_for_launch(self.job)

        # 2) Loop until the job has finished
        start_time = time.time()
        is_running = True
        next_log_update = start_time
        log_update_delay = 1

        while is_running:
            if time.time() > next_log_update:
                _print_available(stdout_tail, 'stdout')
                _print_available(stderr_tail, 'stderr')
                now = time.time()
                log_update_delay = update_delay(now - start_time)
                next_log_update = now + log_update_delay
                is_running = self.job.is_running

            # This sleep should never delay log updates. On the other hand,
            # we should exit this loop when the task has finished without
            # a long delay, regardless of the log tailing schedule
            d = min(log_update_delay, 5.0)
            select.poll().poll(d * 1000)
        
        # 3) Fetch remaining logs
        #
        # It is possible that we exit the loop above before all logs have been
        # shown.
        #
        # TODO if we notice AWS Batch failing to upload logs to S3, we can add a
        # HEAD request here to ensure that the file exists prior to calling
        # S3Tail and note the user about truncated logs if it doesn't
        _print_available(stdout_tail, 'stdout')
        _print_available(stderr_tail, 'stderr')
        # In case of hard crashes (OOM), the final save_logs won't happen.
        # We fetch the remaining logs from AWS CloudWatch and persist them to 
        # Amazon S3.
        #
        # TODO: AWS CloudWatch fetch logs

        if self.job.is_crashed:
            msg = next(msg for msg in 
                [self.job.reason, self.job.status_reason, 'Task crashed.']
                 if msg is not None)
            raise BatchException(
                '%s '
                'This could be a transient error. '
                'Use @retry to retry.' % msg
            )
        else:
            if self.job.is_running:
                # Kill the job if it is still running by throwing an exception.
                raise BatchException("Task failed!")
            echo(
                'Task finished with exit code %s.' % self.job.status_code,
                'stderr',
                batch_id=self.job.id
            )



# NOTE: This file is an example of an unbounded-foreach implementation and is
# used to test the functionality. This is an example only and is not part of the
# core unbounded-foreach implementation.

import os
import subprocess
import sys
from metaflow.cli_args import cli_args
from metaflow.decorators import StepDecorator
from metaflow.exception import MetaflowException
from metaflow.unbounded_foreach import UnboundedForeachInput, UBF_CONTROL, UBF_TASK
from metaflow.util import to_unicode

class InternalTestUnboundedForeachInput(UnboundedForeachInput):
    """
    Test class that wraps around values (any iterator) and simulates an
    unbounded-foreach instead of a bounded foreach.
    """
    NAME = 'InternalTestUnboundedForeachInput'

    def __init__(self, iterable):
        self.iterable = iterable
        super(InternalTestUnboundedForeachInput, self).__init__()

    def __iter__(self):
        return iter(self.iterable)

    def __next__(self):
        return next(self.iter)

    def __getitem__(self, key):
        # Add this for the sake of control task.
        if key is None:
            return self
        return self.iterable[key]

    def __len__(self):
        return len(self.iterable)

    def __str__(self):
        return str(self.iterable)

    def __repr__(self):
        return '%s(%s)' % (self.NAME, self.iterable)

class InternalTestUnboundedForeachDecorator(StepDecorator):
    name = 'unbounded_test_foreach_internal'
    results_dict = {}

    def __init__(self,
                 attributes=None,
                 statically_defined=False):
        super(InternalTestUnboundedForeachDecorator, self).__init__(
            attributes, statically_defined)

    def step_init(self,
                  flow,
                  graph,
                  step_name,
                  decorators,
                  environment,
                  flow_datastore,
                  logger):
        self.environment = environment

    def control_task_step_func(self, flow, graph, retry_count):
        from metaflow import current
        run_id = current.run_id
        step_name = current.step_name
        control_task_id = current.task_id
        (_, split_step_name, split_task_id) = control_task_id.split('-')[1:]

        executable = self.environment.executable(step_name)
        script = sys.argv[0]

        # Access the `unbounded_foreach` param using `flow` (as datastore).
        assert(flow._unbounded_foreach)
        foreach_iter = flow.input
        if not isinstance(foreach_iter, InternalTestUnboundedForeachInput):
            raise MetaflowException('Expected type to be '\
                                    'InternalTestUnboundedForeachInput. Found %s'\
                                    % (type(foreach_iter)))
        foreach_num_splits = sum(1 for _ in foreach_iter)

        print('Simulating UnboundedForeach over value:',
              foreach_iter, 'num_splits:', foreach_num_splits)
        mapper_tasks = []

        for i in range(foreach_num_splits):
            task_id = \
                '%s-%d' % (control_task_id.replace('control-', 'test-ubf-'), i)
            pathspec = '%s/%s/%s' % (run_id, step_name, task_id)
            mapper_tasks.append(to_unicode(pathspec))
            input_paths = '%s/%s/%s' % (run_id, split_step_name, split_task_id)

            # Override specific `step` kwargs.
            kwargs = cli_args.step_kwargs
            kwargs['split_index'] = str(i)
            kwargs['run_id'] = run_id
            kwargs['task_id'] = task_id
            kwargs['input_paths'] = input_paths
            kwargs['ubf_context'] = UBF_TASK
            kwargs['retry_count'] = 0

            cmd = cli_args.step_command(executable, script, step_name,
                                        step_kwargs=kwargs)
            step_cli = u' '.join(cmd)
            # Print cmdline for execution. Doesn't work without the temporary
            # unicode object while using `print`.
            print(u'[${cwd}] Starting split#{split} with cmd:{cmd}'\
                  .format(cwd=os.getcwd(),
                          split=i,
                          cmd=step_cli))
            output_bytes = subprocess.check_output(cmd)
            output = to_unicode(output_bytes)
            for line in output.splitlines():
                print('[Split#%d] %s' % (i, line))
        # Save the list of (child) mapper task pathspec(s) into a designated
        # artifact `_control_mapper_tasks`.
        flow._control_mapper_tasks = mapper_tasks


    def task_decorate(self,
                      step_func,
                      flow,
                      graph,
                      retry_count,
                      max_user_code_retries,
                      ubf_context):
        if ubf_context == UBF_CONTROL:
            from functools import partial
            return partial(self.control_task_step_func, flow, graph, retry_count)
        else:
            return step_func

    def step_task_retry_count(self):
        # UBF plugins don't want retry for the control task. We signal this
        # intent to the runtime by returning (None, None).
        return None, None


from metaflow.decorators import StepDecorator


class ResourcesDecorator(StepDecorator):
    """
    Step decorator to specify the resources needed when executing this step.
    
    This decorator passes this information along to container orchestrator
    (AWS Batch, Kubernetes, etc.) when requesting resources to execute this
    step.
    
    This decorator is ignored if the execution of the step happens locally.
    
    To use, annotate your step as follows:
    ```
    @resources(cpu=32)
    @step
    def my_step(self):
        ...
    ```
    Parameters
    ----------
    cpu : int
        Number of CPUs required for this step. Defaults to 1
    gpu : int
        Number of GPUs required for this step. Defaults to 0
    memory : int
        Memory size (in MB) required for this step. Defaults to 4096
    shared_memory : int
        The value for the size (in MiB) of the /dev/shm volume for this step.
        This parameter maps to the --shm-size option to docker run .
    """
    name = 'resources'
    defaults = {
        'cpu': '1',
        'gpu': '0',
        'memory': '4096',
        'shared_memory': None
    }


import gzip

import io
import json
import os
import pickle
import tempfile

from hashlib import sha1
from shutil import move

import click

from . import parameters
from .datastore.datastore import TransformableObject
from .exception import MetaflowException
from .metaflow_config import DATATOOLS_LOCALROOT, DATATOOLS_SUFFIX
from .parameters import DeployTimeField, Parameter
from .util import to_unicode

try:
    # python2
    from urlparse import urlparse
except:
    # python3
    from urllib.parse import urlparse

# TODO: This local "client" and the general notion of dataclients should probably
# be moved somewhere else. Putting here to keep this change compact for now
class MetaflowLocalURLException(MetaflowException):
    headline = 'Invalid path'

class MetaflowLocalNotFound(MetaflowException):
    headline = 'Local object not found'

class LocalObject(object):
    """
    This object represents a local object. It is a very thin wrapper
    to allow it to be used in the same way as the S3Object (only as needed
    in this usecase)

    Get or list calls return one or more of LocalObjects.
    """

    def __init__(self, url, path):

        # all fields of S3Object should return a unicode object
        def ensure_unicode(x):
            return None if x is None else to_unicode(x)
        path = ensure_unicode(path)

        self._path = path
        self._url = url

        if self._path:
            try:
                os.stat(self._path)
            except FileNotFoundError:
                self._path = None

    @property
    def exists(self):
        """
        Does this key correspond to an actual file?
        """
        return self._path is not None and os.path.isfile(self._path)

    @property
    def url(self):
        """
        Local location of the object; this is the path prefixed with local://
        """
        return self._url

    @property
    def path(self):
        """
        Path to the local file
        """
        return self._path


class Local(object):
    """
    This class allows you to access the local filesystem in a way similar to the S3 datatools
    client. It is a stripped down version for now and only implements the functionality needed
    for this use case.

    In the future, we may want to allow it to be used in a way similar to the S3() client.
    """

    @staticmethod
    def _makedirs(path):
        try:
            os.makedirs(path)
        except OSError as x:
            if x.errno == 17:
                return
            else:
                raise

    @classmethod
    def get_root_from_config(cls, echo, create_on_absent=True):
        result = DATATOOLS_LOCALROOT
        if result is None:
            from .datastore.local import LocalDataStore
            result = LocalDataStore.get_datastore_root_from_config(echo, create_on_absent)
            result = os.path.join(result, DATATOOLS_SUFFIX)
            if create_on_absent and not os.path.exists(result):
                os.mkdir(result)
        return result

    def __init__(self):
        """
        Initialize a new context for Local file operations. This object is based used as
        a context manager for a with statement.
        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def _path(self, key):
        key = to_unicode(key)
        if key.startswith(u'local://'):
            return key[8:]
        elif key[0] != u'/':
            if current.is_running_flow:
                raise MetaflowLocalURLException(
                    "Specify Local(run=self) when you use Local inside a running "
                    "flow. Otherwise you have to use Local with full "
                    "local:// urls or absolute paths.")
            else:
                raise MetaflowLocalURLException(
                    "Initialize Local with an 'localroot' or 'run' if you don't "
                    "want to specify full local:// urls or absolute paths.")
        else:
            return key

    def get(self, key=None, return_missing=False):
        p = self._path(key)
        url = u'local://%s' % p
        if not os.path.isfile(p):
            if return_missing:
                p = None
            else:
                raise MetaflowLocalNotFound("Local URL %s not found" % url)
        return LocalObject(url, p)

    def put(self, key, obj, overwrite=True):
        p = self._path(key)
        if overwrite or (not os.path.exists(p)):
            Local._makedirs(os.path.dirname(p))
            with open(p, 'wb') as f:
                f.write(obj)
        return u'local://%s' % p


# From here on out, this is the IncludeFile implementation.
from .datatools import S3

DATACLIENTS = {'local': Local,
               's3': S3}

class LocalFile():
    def __init__(self, is_text, encoding, path):
        self._is_text = is_text
        self._encoding = encoding
        self._path = path

    @classmethod
    def is_file_handled(cls, path):
        if path:
            decoded_value = Uploader.decode_value(to_unicode(path))
            if decoded_value['type'] == 'self':
                return True, LocalFile(
                    decoded_value['is_text'], decoded_value['encoding'],
                    decoded_value['url']), None
            path = decoded_value['url']
        for prefix, handler in DATACLIENTS.items():
            if path.startswith(u"%s://" % prefix):
                return True, Uploader(handler), None
        try:
            with open(path, mode='r') as _:
                pass
        except OSError:
            return False, None, "IncludeFile: could not open file '%s'" % path
        return True, None, None

    def __str__(self):
        return self._path

    def __repr__(self):
        return self._path

    def __call__(self, ctx):
        # We check again if this is a local file that exists. We do this here because
        # we always convert local files to DeployTimeFields irrespective of whether
        # the file exists.
        ok, _, err = LocalFile.is_file_handled(self._path)
        if not ok:
            raise MetaflowException(err)
        client = DATACLIENTS.get(ctx.ds_type)
        if client:
            return Uploader(client).store(
                ctx.flow_name, self._path, self._is_text, self._encoding, ctx.logger)
        raise MetaflowException(
            "IncludeFile: no client found for datastore type %s" % ctx.ds_type)


class FilePathClass(click.ParamType):
    name = 'FilePath'
    # The logic for this class is as follows:
    #  - It will always return a path that indicates the persisted path of the file.
    #    + If the value is already such a string, nothing happens and it returns that same value
    #    + If the value is a LocalFile, it will persist the local file and return the path
    #      of the persisted file
    #  - The artifact will be persisted prior to any run (for non-scheduled runs through persist_parameters)
    #    + This will therefore persist a simple string
    #  - When the parameter is loaded again, the load_parameter in the IncludeFile class will get called
    #    which will download and return the bytes of the persisted file.
    def __init__(self, is_text, encoding):
        self._is_text = is_text
        self._encoding = encoding

    def convert(self, value, param, ctx):
        if callable(value):
            # Already a correct type
            return value

        value = os.path.expanduser(value)
        ok, file_type, err = LocalFile.is_file_handled(value)
        if not ok:
            self.fail(err)
        if file_type is None:
            # Here, we need to store the file
            return lambda is_text=self._is_text, encoding=self._encoding,\
                value=value, ctx=parameters.context_proto: LocalFile(
                    is_text, encoding, value)(ctx)
        elif isinstance(file_type, LocalFile):
            # This is a default file that we evaluate now (to delay upload
            # until *after* the flow is checked)
            return lambda f=file_type, ctx=parameters.context_proto: f(ctx)
        else:
            # We will just store the URL in the datastore along with text/encoding info
            return lambda is_text=self._is_text, encoding=self._encoding,\
                value=value: Uploader.encode_url(
                    'external', value, is_text=is_text, encoding=encoding)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return 'FilePath'


class IncludeFile(Parameter):

    def __init__(
            self, name, required=False, is_text=True, encoding=None, help=None, **kwargs):
        # Defaults are DeployTimeField
        v = kwargs.get('default')
        if v is not None:
            _, file_type, _ = LocalFile.is_file_handled(v)
            # Ignore error because we may never use the default
            if file_type is None:
                o = {
                    'type': 'self',
                    'is_text': is_text,
                    'encoding': encoding,
                    'url': v
                }
                kwargs['default'] = DeployTimeField(
                    name,
                    str,
                    'default',
                    lambda ctx, full_evaluation, o=o: \
                    LocalFile(o['is_text'], o['encoding'], o['url'])(ctx) \
                        if full_evaluation else json.dumps(o),
                    print_representation=v)
            else:
                kwargs['default'] = DeployTimeField(
                    name,
                    str,
                    'default',
                    lambda _, __, is_text=is_text, encoding=encoding, v=v: \
                        Uploader.encode_url('external-default', v,
                                            is_text=is_text, encoding=encoding),
                    print_representation=v)

        super(IncludeFile, self).__init__(
            name, required=required, help=help,
            type=FilePathClass(is_text, encoding), **kwargs)

    def load_parameter(self, val):
        if val is None:
            return val
        ok, file_type, err = LocalFile.is_file_handled(val)
        if not ok:
            raise MetaflowException("Parameter '%s' could not be loaded: %s" % (self.name, err))
        if file_type is None or isinstance(file_type, LocalFile):
            raise MetaflowException("Parameter '%s' was not properly converted" % self.name)
        return file_type.load(val)


class Uploader():

    file_type = 'uploader-v1'

    def __init__(self, client_class):
        self._client_class = client_class

    @staticmethod
    def encode_url(url_type, url, **kwargs):
        # Avoid encoding twice (default -> URL -> _convert method of FilePath for example)
        if url is None or len(url) == 0 or url[0] == '{':
            return url
        return_value = {'type': url_type, 'url': url}
        return_value.update(kwargs)
        return json.dumps(return_value)

    @staticmethod
    def decode_value(value):
        if value is None or len(value) == 0 or value[0] != '{':
            return {'type': 'base', 'url': value}
        return json.loads(value)

    def store(self, flow_name, path, is_text, encoding, echo):
        sz = os.path.getsize(path)
        unit = ['B', 'KB', 'MB', 'GB', 'TB']
        pos = 0
        while pos < len(unit) and sz >= 1024:
            sz = sz // 1024
            pos += 1
        if pos >= 3:
            extra = '(this may take a while)'
        else:
            extra = ''
        echo(
            'Including file %s of size %d%s %s' % (path, sz, unit[pos], extra))
        try:
            cur_obj = TransformableObject(io.open(path, mode='rb').read())
        except IOError:
            # If we get an error here, since we know that the file exists already,
            # it means that read failed which happens with Python 2.7 for large files
            raise MetaflowException('Cannot read file at %s -- this is likely because it is too '
                                    'large to be properly handled by Python 2.7' % path)
        sha = sha1(cur_obj.current()).hexdigest()
        path = os.path.join(self._client_class.get_root_from_config(echo, True),
                            flow_name,
                            sha)
        buf = io.BytesIO()
        with gzip.GzipFile(
                fileobj=buf, mode='wb', compresslevel=3) as f:
            f.write(cur_obj.current())
        cur_obj.transform(lambda _: buf)
        buf.seek(0)
        with self._client_class() as client:
            url = client.put(path, buf.getvalue(), overwrite=False)
            echo('File persisted at %s' % url)
            return Uploader.encode_url(Uploader.file_type, url, is_text=is_text, encoding=encoding)

    def load(self, value):
        value_info = Uploader.decode_value(value)
        with self._client_class() as client:
            obj = client.get(value_info['url'], return_missing=True)
            if obj.exists:
                if value_info['type'] == Uploader.file_type:
                    # We saved this file directly so we know how to read it out
                    with gzip.GzipFile(filename=obj.path, mode='rb') as f:
                        if value_info['is_text']:
                            return io.TextIOWrapper(f, encoding=value_info.get('encoding')).read()
                        return f.read()
                else:
                    # We open this file according to the is_text and encoding information
                    if value_info['is_text']:
                        return io.open(obj.path, mode='rt', encoding=value_info.get('encoding')).read()
                    else:
                        return io.open(obj.path, mode='rb').read()
            raise FileNotFoundError("File at %s does not exist" % value_info['url'])



import click
from metaflow.package import MetaflowPackage

@click.group()
def cli():
    pass

@cli.group(help='Commands related to code packages.')
@click.pass_obj
def package(obj):
    # Prepare the package before any of the sub-commands are invoked.
    obj.package = MetaflowPackage(obj.flow,
                                  obj.environment,
                                  obj.echo,
                                  obj.package_suffixes)

@package.command(help='Output information about the current code package.')
@click.pass_obj
def info(obj):
    obj.echo('Status of the current working directory:', fg='magenta', bold=False)
    obj.echo_always('Hash: *%s*' % obj.package.sha,
                    highlight='green',
                    highlight_bold=False)
    obj.echo_always('Package size: *%d* KB' % (len(obj.package.blob) / 1024),
                    highlight='green',
                    highlight_bold=False)
    num = sum(1 for _ in obj.package.path_tuples())
    obj.echo_always('Number of files: *%d*' % num,
                    highlight='green',
                    highlight_bold=False)

@package.command(help='List files included in the code package.')
@click.pass_obj
def list(obj):
    obj.echo('Files included in the code package '
             '(change with --package-suffixes):',
             fg='magenta',
             bold=False)
    obj.echo_always('\n'.join(path for path, _ in obj.package.path_tuples()))

@package.command(help='Save the current code package in a tar file')
@click.argument('path')
@click.pass_obj
def save(obj, path):
    with open(path, 'wb') as f:
        f.write(obj.package.blob)
    obj.echo('Code package saved in *%s*.' % path,
             fg='magenta',
             bold=False)



import os
import requests
import time

from distutils.version import LooseVersion

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import METADATA_SERVICE_NUM_RETRIES, METADATA_SERVICE_HEADERS, \
    METADATA_SERVICE_URL
from metaflow.metadata import MetadataProvider
from metaflow.metadata.heartbeat import HB_URL_KEY
from metaflow.sidecar import SidecarSubProcess
from metaflow.sidecar_messages import MessageTypes, Message

# Define message enums
class HeartbeatTypes(object):
    RUN = 1
    TASK = 2

class ServiceException(MetaflowException):
    headline = 'Metaflow service error'

    def __init__(self, msg, http_code=None, body=None):
        self.http_code = None if http_code is None else int(http_code)
        self.response = body
        super(ServiceException, self).__init__(msg)


class ServiceMetadataProvider(MetadataProvider):
    TYPE = 'service'

    def __init__(self, environment, flow, event_logger, monitor):
        super(ServiceMetadataProvider, self).__init__(environment, flow, event_logger, monitor)
        self.url_task_template = os.path.join(METADATA_SERVICE_URL,
                                             'flows/{flow_id}/runs/{run_number}/steps/{step_name}/tasks/{task_id}/heartbeat')
        self.url_run_template = os.path.join(METADATA_SERVICE_URL,
                                            'flows/{flow_id}/runs/{run_number}/heartbeat')
        self.sidecar_process = None

    @classmethod
    def compute_info(cls, val):
        v = val.rstrip('/')
        try:
            resp = requests.get(os.path.join(v, 'ping'), headers=METADATA_SERVICE_HEADERS)
            resp.raise_for_status()
        except:  # noqa E722
            raise ValueError('Metaflow service [%s] unreachable.' % v)
        return v

    @classmethod
    def default_info(cls):
        return METADATA_SERVICE_URL

    def version(self):
        return self._version(self._monitor)

    def new_run_id(self, tags=[], sys_tags=[]):
        return self._new_run(tags=tags, sys_tags=sys_tags)

    def register_run_id(self, run_id, tags=[], sys_tags=[]):
        try:
            # don't try to register an integer ID which was obtained
            # from the metadata service in the first place
            int(run_id)
            return
        except ValueError:
            return self._new_run(run_id, tags=tags, sys_tags=sys_tags)

    def new_task_id(self, run_id, step_name, tags=[], sys_tags=[]):
        return self._new_task(run_id, step_name, tags=tags, sys_tags=sys_tags)

    def register_task_id(self,
                         run_id,
                         step_name,
                         task_id,
                         tags=[],
                         sys_tags=[]):
        try:
            # don't try to register an integer ID which was obtained
            # from the metadata service in the first place
            int(task_id)
        except ValueError:
            self._new_task(run_id,
                           step_name,
                           task_id,
                           tags=tags,
                           sys_tags=sys_tags)
        finally:
            self._register_code_package_metadata(run_id, step_name, task_id)

    def _start_heartbeat(self, heartbeat_type, flow_id, run_id, step_name=None, task_id=None):
        if self._already_started():
            # A single ServiceMetadataProvider instance can not start
            # multiple heartbeat side cars of any type/combination. Either a
            # single run heartbeat or a single task heartbeat can be started
            raise Exception("heartbeat already started")
        # start sidecar
        if self.version() is None or \
                LooseVersion(self.version()) < LooseVersion('2.0.4'):
            # if old version of the service is running
            # then avoid running real heartbeat sidecar process
            self.sidecar_process = SidecarSubProcess("nullSidecarHeartbeat")
        else:
            self.sidecar_process = SidecarSubProcess("heartbeat")
        # create init message
        payload = {}
        if heartbeat_type == HeartbeatTypes.TASK:
            # create task heartbeat
            data = {
                    'flow_id': flow_id, 'run_number': run_id,
                    'step_name': step_name, 'task_id': task_id,
                    }
            payload[HB_URL_KEY] = self.url_task_template.format(**data)
        elif heartbeat_type == HeartbeatTypes.RUN:
            # create run heartbeat
            data = {'flow_id': flow_id, 'run_number': run_id}

            payload[HB_URL_KEY] = self.url_run_template.format(**data)
        else:
            raise Exception("invalid heartbeat type")
        payload["service_version"] = self.version()
        msg = Message(MessageTypes.LOG_EVENT, payload)
        self.sidecar_process.msg_handler(msg)

    def start_run_heartbeat(self, flow_id, run_id):
        self._start_heartbeat(HeartbeatTypes.RUN, flow_id, run_id)

    def start_task_heartbeat(self, flow_id, run_id, step_name, task_id):
        self._start_heartbeat(HeartbeatTypes.TASK,
                              flow_id,
                              run_id,
                              step_name,
                              task_id)

    def _already_started(self):
        return self.sidecar_process is not None

    def stop_heartbeat(self):
        msg = Message(MessageTypes.SHUTDOWN, None)
        self.sidecar_process.msg_handler(msg)

    def register_data_artifacts(self,
                                run_id,
                                step_name,
                                task_id,
                                attempt_id,
                                artifacts):
        url = ServiceMetadataProvider._obj_path(self._flow_name, run_id, step_name, task_id)
        url += '/artifact'
        data = self._artifacts_to_json(run_id, step_name, task_id, attempt_id, artifacts)
        self._request(self._monitor, url, data)

    def register_metadata(self, run_id, step_name, task_id, metadata):
        url = ServiceMetadataProvider._obj_path(self._flow_name, run_id, step_name, task_id)
        url += '/metadata'
        data = self._metadata_to_json(run_id, step_name, task_id, metadata)
        self._request(self._monitor, url, data)

    @classmethod
    def _get_object_internal(cls, obj_type, obj_order, sub_type, sub_order, filters=None, *args):
        # Special handling of self, artifact, and metadata
        if sub_type == 'self':
            url = ServiceMetadataProvider._obj_path(*args[:obj_order])
            try:
                return MetadataProvider._apply_filter([cls._request(None, url)], filters)[0]
            except ServiceException as ex:
                if ex.http_code == 404:
                    return None
                raise

        # For the other types, we locate all the objects we need to find and return them
        if obj_type != 'root':
            url = ServiceMetadataProvider._obj_path(*args[:obj_order])
        else:
            url = ''
        if sub_type != 'metadata':
            url += '/%ss' % sub_type
        else:
            url += '/metadata'
        try:
            return MetadataProvider._apply_filter(cls._request(None, url), filters)
        except ServiceException as ex:
            if ex.http_code == 404:
                return None
            raise

    def _new_run(self, run_id=None, tags=[], sys_tags=[]):
        # first ensure that the flow exists
        self._get_or_create('flow')
        run = self._get_or_create('run', run_id, tags=tags, sys_tags=sys_tags)
        return str(run['run_number'])

    def _new_task(self,
                  run_id,
                  step_name,
                  task_id=None,
                  tags=[],
                  sys_tags=[]):
        # first ensure that the step exists
        self._get_or_create('step', run_id, step_name)
        task = self._get_or_create('task', run_id, step_name, task_id, tags=tags, sys_tags=sys_tags)
        self._register_code_package_metadata(run_id, step_name, task['task_id'])
        return task['task_id']

    @staticmethod
    def _obj_path(
            flow_name, run_id=None, step_name=None, task_id=None, artifact_name=None):
        object_path = '/flows/%s' % flow_name
        if run_id:
            object_path += '/runs/%s' % run_id
        if step_name:
            object_path += '/steps/%s' % step_name
        if task_id:
            object_path += '/tasks/%s' % task_id
        if artifact_name:
            object_path += '/artifacts/%s' % artifact_name
        return object_path

    @staticmethod
    def _create_path(obj_type, flow_name, run_id=None, step_name=None):
        create_path = '/flows/%s' % flow_name
        if obj_type == 'flow':
            return create_path
        if obj_type == 'run':
            return create_path + '/run'
        create_path += '/runs/%s/steps/%s' % (run_id, step_name)
        if obj_type == 'step':
            return create_path + '/step'
        return create_path + '/task'

    def _get_or_create(
            self, obj_type, run_id=None, step_name=None, task_id=None, tags=[], sys_tags=[]):

        def create_object():
            data = self._object_to_json(
                obj_type,
                run_id,
                step_name,
                task_id,
                tags + self.sticky_tags,
                sys_tags + self.sticky_sys_tags)
            return self._request(self._monitor, create_path, data, obj_path)

        always_create = False
        obj_path = self._obj_path(self._flow_name, run_id, step_name, task_id)
        create_path = self._create_path(obj_type, self._flow_name, run_id, step_name)
        if obj_type == 'run' and run_id is None:
            always_create = True
        elif obj_type == 'task' and task_id is None:
            always_create = True

        if always_create:
            return create_object()

        try:
            return self._request(self._monitor, obj_path)
        except ServiceException as ex:
            if ex.http_code == 404:
                return create_object()
            else:
                raise

    @classmethod
    def _request(cls, monitor, path, data=None, retry_409_path=None):
        if cls.INFO is None:
            raise MetaflowException('Missing Metaflow Service URL. '
                'Specify with METAFLOW_SERVICE_URL environment variable')
        url = os.path.join(cls.INFO, path.lstrip('/'))
        for i in range(METADATA_SERVICE_NUM_RETRIES):
            try:
                if data is None:
                    if monitor:
                        with monitor.measure('metaflow.service_metadata.get'):
                            resp = requests.get(url, headers=METADATA_SERVICE_HEADERS)
                    else:
                        resp = requests.get(url, headers=METADATA_SERVICE_HEADERS)
                else:
                    if monitor:
                        with monitor.measure('metaflow.service_metadata.post'):
                            resp = requests.post(url, headers=METADATA_SERVICE_HEADERS, json=data)
                    else:
                        resp = requests.post(url, headers=METADATA_SERVICE_HEADERS, json=data)
            except:  # noqa E722
                if monitor:
                    with monitor.count('metaflow.service_metadata.failed_request'):
                        if i == METADATA_SERVICE_NUM_RETRIES - 1:
                            raise
                else:
                    if i == METADATA_SERVICE_NUM_RETRIES - 1:
                        raise
                resp = None
            else:
                if resp.status_code < 300:
                    return resp.json()
                elif resp.status_code == 409 and data is not None:
                    # a special case: the post fails due to a conflict
                    # this could occur when we missed a success response
                    # from the first POST request but the request
                    # actually went though, so a subsequent POST
                    # returns 409 (conflict) or we end up with a
                    # conflict while running on AWS Step Functions
                    # instead of retrying the post we retry with a get since
                    # the record is guaranteed to exist
                    if retry_409_path:
                        return cls._request(monitor, retry_409_path)
                    else:
                        return
                elif resp.status_code != 503:
                    raise ServiceException('Metadata request (%s) failed (code %s): %s'
                                           % (path, resp.status_code, resp.text),
                                           resp.status_code,
                                           resp.text)
            time.sleep(2**i)

        if resp:
            raise ServiceException('Metadata request (%s) failed (code %s): %s'
                                   % (path, resp.status_code, resp.text),
                                   resp.status_code,
                                   resp.text)
        else:
            raise ServiceException('Metadata request (%s) failed' % path)

    @classmethod
    def _version(cls, monitor):
        if cls.INFO is None:
            raise MetaflowException('Missing Metaflow Service URL. '
                'Specify with METAFLOW_SERVICE_URL environment variable')
        path = 'ping'
        url = os.path.join(cls.INFO, path)
        for i in range(METADATA_SERVICE_NUM_RETRIES):
            try:
                if monitor:
                    with monitor.measure('metaflow.service_metadata.get'):
                        resp = requests.get(url,
                            headers=METADATA_SERVICE_HEADERS)
                else:
                    resp = requests.get(url, headers=METADATA_SERVICE_HEADERS)
            except:
                if monitor:
                    with monitor.count(
                        'metaflow.service_metadata.failed_request'):
                        if i == METADATA_SERVICE_NUM_RETRIES - 1:
                            raise
                else:
                    if i == METADATA_SERVICE_NUM_RETRIES - 1:
                        raise
                resp = None
            else:
                if resp.status_code < 300:
                    return resp.headers.get('METADATA_SERVICE_VERSION', None)
                elif resp.status_code != 503:
                    raise ServiceException('Metadata request (%s) failed'
                                           ' (code %s): %s' %
                                           (path, resp.status_code, resp.text),
                                           resp.status_code,
                                           resp.text)
            time.sleep(2**i)
        if resp:
            raise ServiceException('Metadata request (%s) failed (code %s): %s'
                                   % (path, resp.status_code, resp.text),
                                   resp.status_code,
                                   resp.text)
        else:
            raise ServiceException('Metadata request (%s) failed' % path)



import os
import imp
from tempfile import NamedTemporaryFile

from .util import to_bytes

R_FUNCTIONS = {}
R_PACKAGE_PATHS = None
RDS_FILE_PATH = None
R_CONTAINER_IMAGE = None
METAFLOW_R_VERSION = None
R_VERSION = None
R_VERSION_CODE = None

def call_r(func_name, args):
    R_FUNCTIONS[func_name](*args)

def get_r_func(func_name):
    return R_FUNCTIONS[func_name]

def package_paths():
    if R_PACKAGE_PATHS is not None:
        root = R_PACKAGE_PATHS['package']
        prefixlen = len('%s/' % root.rstrip('/'))
        for path, dirs, files in os.walk(R_PACKAGE_PATHS['package']):
            if '/.' in path:
                continue
            for fname in files:
                if fname[0] == '.':
                    continue
                p = os.path.join(path, fname)
                yield p, os.path.join('metaflow-r', p[prefixlen:])
        flow = R_PACKAGE_PATHS['flow']
        yield flow, os.path.basename(flow)

def entrypoint():
    return 'PYTHONPATH=/root/metaflow R_LIBS_SITE=`Rscript -e \'cat(paste(.libPaths(), collapse=\\":\\"))\'`:metaflow/ Rscript metaflow-r/run_batch.R --flowRDS=%s' % RDS_FILE_PATH

def use_r():
    return R_PACKAGE_PATHS is not None

def container_image():
    return R_CONTAINER_IMAGE

def metaflow_r_version():
    return METAFLOW_R_VERSION

def r_version():
    return R_VERSION

def r_version_code():
    return R_VERSION_CODE

def working_dir():
    if use_r(): 
        return R_PACKAGE_PATHS['wd']
    return None

def run(flow_script,
        r_functions,
        rds_file,
        metaflow_args,
        full_cmdline,
        r_paths,
        r_container_image,
        metaflow_r_version,
        r_version,
        r_version_code):
    global R_FUNCTIONS, \
        R_PACKAGE_PATHS, \
        RDS_FILE_PATH, \
        R_CONTAINER_IMAGE, \
        METAFLOW_R_VERSION, \
        R_VERSION, \
        R_VERSION_CODE

    R_FUNCTIONS = r_functions
    R_PACKAGE_PATHS = r_paths
    RDS_FILE_PATH = rds_file
    R_CONTAINER_IMAGE = r_container_image
    METAFLOW_R_VERSION = metaflow_r_version
    R_VERSION = r_version
    R_VERSION_CODE = r_version_code

    # there's some reticulate(?) sillyness which causes metaflow_args
    # not to be a list if it has only one item. Here's a workaround
    if not isinstance(metaflow_args, list):
        metaflow_args = [metaflow_args]
    # remove any reference to local path structure from R
    full_cmdline[0] = os.path.basename(full_cmdline[0])
    with NamedTemporaryFile(prefix="metaflowR.", delete=False) as tmp:
        tmp.write(to_bytes(flow_script))
    module = imp.load_source('metaflowR', tmp.name)
    flow = module.FLOW(use_cli=False)

    from . import exception 
    from . import cli 
    try:
        cli.main(flow,
                 args=metaflow_args,
                 handle_exceptions=False,
                 entrypoint=full_cmdline[:-len(metaflow_args)])
    except exception.MetaflowException as e:
        cli.print_metaflow_exception(e)
        os.remove(tmp.name)
        os._exit(1)
    except Exception as e:
        import sys
        print(e)
        sys.stdout.flush()
        os.remove(tmp.name)
        os._exit(1)
    finally:
        os.remove(tmp.name)


"""
Welcome to Metaflow!

Metaflow is a microframework for data science projects.

There are two main use cases for this package:

1) You can define new flows using the `FlowSpec`
   class and related decorators.

2) You can access and inspect existing flows.
   You can instantiate a `Metaflow` class to
   get an entry point to all existing objects.

# How to work with flows

A flow is a directed graph of Python functions called steps.
Metaflow takes care of executing these steps one by one in various
environments, such as on a local laptop or compute environments
(such as AWS Batch for example). It snapshots
data and code related to each run, so you can resume, reproduce,
and inspect results easily at a later point in time.

Here is a high-level overview of objects related to flows:

 [ FlowSpec ]     (0) Base class for flows.
 [ MyFlow ]       (1) Subclass from FlowSpec to define a new flow.

define new flows
----------------- (2) Run MyFlow on the command line.
access results

 [ Flow ]         (3) Access your flow with `Flow('MyFlow')`.
 [ Run ]          (4) Access a specific run with `Run('MyFlow/RunID')`.
 [ Step ]         (5) Access a specific step by its name, e.g. `run['end']`.
 [ Task ]         (6) Access a task related to the step with `step.task`.
 [ DataArtifact ] (7) Access data of a task with `task.data`.

# More questions?

If you have any questions, feel free to post a bug report/question on the
Metaflow Github page.
"""

import importlib
import sys
import types

if sys.version_info[0] >= 3 and sys.version_info[1] >= 4:
    import importlib.util
    from importlib.machinery import ModuleSpec
else:
    # Something random so there is no syntax error
    ModuleSpec = None

class _LazyLoader(object):
    # This _LazyLoader implements the Importer Protocol defined in PEP 302
    def __init__(self, handled):
        # Modules directly loaded (this is either new modules or overrides of existing ones)
        self._handled = handled if handled else {}

        # This is used to revert back to regular loading when trying to load
        # the over-ridden module
        self._tempexcluded = set()

    def find_module(self, fullname, path=None):
        if fullname in self._tempexcluded:
            return None
        if fullname in self._handled or \
                (fullname.endswith('._orig') and fullname[:-6] in self._handled):
            return self
        return None

    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]
        if not self._can_handle_orig_module() and fullname.endswith('._orig'):
            # We return a nicer error message
            raise ImportError(
                "Attempting to load '%s' -- loading shadowed modules in Metaflow "
                "Custom is only supported in Python 3.4+" % fullname)
        to_import = self._handled.get(fullname, None)
        # We see if we are shadowing an existing module and, if so, we
        # will keep track of the module we are shadowing so that it
        # may be loaded if needed. We basically will create a ._orig submodule
        # of sorts. This functionality only works for Python 3.4+. For anything
        # below this, we do not create the _orig module so loading it will
        # result in ModuleNotFound
        if self._can_handle_orig_module() and not fullname.endswith('._orig'):
            try:
                # We exclude this module temporarily from what we handle to
                # revert back to the non-shadowing mode of import
                self._tempexcluded.add(fullname)
                spec = importlib.util.find_spec(fullname)
                self._handled["%s._orig" % fullname] = spec
            finally:
                self._tempexcluded.remove(fullname)

        if isinstance(to_import, str):
            to_import = importlib.import_module(to_import)
            sys.modules[fullname] = to_import
        elif isinstance(to_import, types.ModuleType):
            sys.modules[fullname] = to_import
        elif self._can_handle_orig_module() and isinstance(to_import, ModuleSpec):
            # This loads modules that end in _orig
            m = importlib.util.module_from_spec(to_import)
            to_import.loader.exec_module(m)
            sys.modules[fullname] = m
        elif to_import is None and fullname.endswith('._orig'):
            # This happens when trying to access a shadowed ._orig module
            # when actually, there is no shadowed module; print a nicer message
            # Condition is a bit overkill and most likely only checking to_import
            # would be OK. Being extra sure in case _LazyLoader is misused and
            # a None value is passed in.
            raise ImportError(
                "Metaflow Custom shadowed module '%s' does not exist" % fullname)
        else:
            raise ImportError
        return sys.modules[fullname]

    @staticmethod
    def _can_handle_orig_module():
        return sys.version_info[0] >= 3 and sys.version_info[1] >= 4


from .event_logger import EventLogger

# Flow spec
from .flowspec import FlowSpec
from .includefile import IncludeFile
from .parameters import Parameter, JSONTypeClass
JSONType = JSONTypeClass()

# current runtime singleton
from .current import current

# data layer
from .datatools import S3

# Decorators
from .decorators import step, _import_plugin_decorators
# this auto-generates decorator functions from Decorator objects
# in the top-level metaflow namespace
_import_plugin_decorators(globals())

# Client
from .client import namespace,\
                    get_namespace,\
                    default_namespace,\
                    metadata, \
                    get_metadata, \
                    default_metadata, \
                    Metaflow,\
                    Flow,\
                    Run,\
                    Step,\
                    Task,\
                    DataArtifact

# Utilities
from .multicore_utils import parallel_imap_unordered,\
                             parallel_map
from .metaflow_profile import profile


__version_addl__ = None
try:
    import metaflow_custom.toplevel as extension_module
except ImportError as e:
    ver = sys.version_info[0] * 10 + sys.version_info[1]
    if ver >= 36:
        # e.name is set to the name of the package that fails to load
        # so don't error ONLY IF the error is importing this module (but do
        # error if there is a transitive import error)
        if not (isinstance(e, ModuleNotFoundError) and \
                e.name in ['metaflow_custom', 'metaflow_custom.toplevel']):
            print(
                "Cannot load metaflow_custom top-level configuration -- "
                "if you want to ignore, uninstall metaflow_custom package")
            raise
else:
    # We load into globals whatever we have in extension_module
    # We specifically exclude any modules that may be included (like sys, os, etc)
    # *except* for ones that are part of metaflow_custom (basically providing
    # an aliasing mechanism)
    lazy_load_custom_modules = {}
    addl_modules = extension_module.__dict__.get('__mf_promote_submodules__')
    if addl_modules:
        # We make an alias for these modules which the metaflow_custom author
        # wants to expose but that may not be loaded yet
        lazy_load_custom_modules = {
            'metaflow.%s' % k: 'metaflow_custom.%s' % k for k in addl_modules}
    for n, o in extension_module.__dict__.items():
        if not n.startswith('__') and not isinstance(o, types.ModuleType):
            globals()[n] = o
        elif isinstance(o, types.ModuleType) and o.__package__ and \
                o.__package__.startswith('metaflow_custom'):
            lazy_load_custom_modules['metaflow.%s' % n] = o
    if lazy_load_custom_modules:
        # Prepend to make sure custom package overrides things
        sys.meta_path = [_LazyLoader(lazy_load_custom_modules)] + sys.meta_path
    __version_addl__ = getattr(extension_module, '__mf_customization__', '<unk>')
    if extension_module.__version__:
        __version_addl__ = '%s(%s)' % (__version_addl__, extension_module.__version__)
finally:
    # Erase all temporary names to avoid leaking things
    for _n in ['ver', 'n', 'o', 'e', 'lazy_load_custom_modules',
               'extension_module', 'addl_modules']:
        try:
            del globals()[_n]
        except KeyError:
            pass
    del globals()['_n']

import pkg_resources
try:
    __version__ = pkg_resources.get_distribution('metaflow').version
except:
    # this happens on remote environments since the job package
    # does not have a version
    __version__ = None



# Portions of this plugin are inspired by RPyC
# https://rpyc.readthedocs.io/
#
# The license for that software is reproduced below
#
# Copyright (c) 2005-2013
#  Tomer Filiba (tomerfiliba@gmail.com)
#  Copyrights of patches are held by their respective submitters
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import pickle
import sys
from subprocess import Popen, PIPE

from itertools import chain

from .exception_transferer import RemoteInterpreterException
from .client_modules import create_modules

# Determine what is the python executable to use for the environment escape. To do this,
# we look for ENV_ESCAPE_PY in the environment AND store it. When metaflow
# is first launched by the user, it is running outside any metaflow created
# environment (like the ones created through the Conda plugin) and we will therefore
# consider that as the environment we escape to.
# Note that it is important to store the value back in the environment to make
# it available to any sub-process that launch sa well.
# We also store the maximum protocol version that we support for pickle so that
# we can determine what to use
ENV_ESCAPE_PY = os.environ.get('METAFLOW_ENV_ESCAPE_PY', sys.executable)
ENV_ESCAPE_PICKLE_VERSION = os.environ.get(
    'METAFLOW_ENV_ESCAPE_PICKLE_VERSION', str(pickle.HIGHEST_PROTOCOL))
os.environ['METAFLOW_ENV_ESCAPE_PICKLE_VERSION'] = ENV_ESCAPE_PICKLE_VERSION
os.environ['METAFLOW_ENV_ESCAPE_PY'] = ENV_ESCAPE_PY


def generate_trampolines(python_path):
    # This function will look in the configurations directory and create
    # files named <module>.py that will properly setup the environment escape when
    # called

    # in some cases we may want to disable environment escape
    # functionality, in that case, set METAFLOW_ENV_ESCAPE_DISABLED
    if os.environ.get('METAFLOW_ENV_ESCAPE_DISABLED', False) in (True, 'True'):
        return

    python_interpreter_path = ENV_ESCAPE_PY
    max_pickle_version = int(ENV_ESCAPE_PICKLE_VERSION)

    paths = [os.path.dirname(os.path.abspath(__file__)) + "/configurations"]
    try:
        import metaflow_custom.plugins.env_escape as custom_escape
    except ImportError as e:
        ver = sys.version_info[0] * 10 + sys.version_info[1]
        if ver >= 36:
            # e.name is set to the name of the package that fails to load
            # so don't error ONLY IF the error is importing this module (but do
            # error if there is a transitive import error)
            if not (isinstance(e, ModuleNotFoundError) and e.name in [
                    'metaflow_custom', 'metaflow_custom.plugins',
                    'metaflow_custom.plugins.env_escape']):
                print(
                    "Cannot load metaflow_custom env escape configurations -- "
                    "if you want to ignore, uninstall metaflow_custom package")
                raise
    else:
        paths.append(os.path.dirname(os.path.abspath(custom_escape.__file__)) +
                     "/configurations")

    for rootpath in paths:
        for path in os.listdir(rootpath):
            path = os.path.join(rootpath, path)
            if os.path.isdir(path):
                dir_name = os.path.basename(path)
                if dir_name.startswith('emulate_'):
                    module_names = dir_name[8:].split("__")
                    for module_name in module_names:
                        with open(os.path.join(
                                python_path, module_name + ".py"), mode='w') as f:
                            f.write("""
import importlib
import os
import sys
from metaflow.plugins.env_escape.client_modules import ModuleImporter

# This is a trampoline file to ensure that the ModuleImporter to handle the emulated
# modules gets properly loaded. If multiple modules are emulated by a single configuration
# the first import will cause this trampoline file to be loaded. All other calls will use
# the module importer since we insert the importer at the head of the meta_path

def load():
    # We check if we are not overriding a module that already exists; to do so, we remove
    # the path we live at from sys.path and check
    old_paths = sys.path
    cur_path = os.path.dirname(__file__)
    sys.path = [p for p in old_paths if p != cur_path]
    # Handle special case where we launch a shell (including with a command)
    # and we are in the CWD (searched if '' is the first element of sys.path)
    if cur_path == os.getcwd() and sys.path[0] == '':
        sys.path = sys.path[1:]

    # Remove the module (this file) to reload it properly. Do *NOT* update sys.modules but
    # modify directly since it may be referenced elsewhere
    del sys.modules["{module_name}"]

    for prefix in {prefixes}:
        try:
            importlib.import_module(prefix)
        except ImportError:
            # Here we actually have two cases: we are being imported from the client (inner env)
            # in which case we are happy (since no module exists) OR we are being imported by the
            # server in which case we could not find the underlying module so we re-raise
            # this error.
            # We distinguish these cases by checking if the executable is the python_path the
            # server should be using
            if sys.executable == "{python_path}":
                raise
            # print("Env escape using executable {python_path}")
        else:
            # Inverse logic as above here.
            if sys.executable == "{python_path}":
                return
            raise RuntimeError("Trying to override '%s' when module exists in system" % prefix)
    sys.path = old_paths
    m = ModuleImporter("{python_path}", {max_pickle_version}, "{path}", {prefixes})
    sys.meta_path.insert(0, m)
    # Reload this module using the ModuleImporter
    importlib.import_module("{module_name}")

if not "{python_path}":
    raise RuntimeError(
        "Trying to access an escaped module ({module_name}) without a valid interpreter")
load()
""" .format(
    python_path=python_interpreter_path,
    max_pickle_version=max_pickle_version,
    path=path,
    prefixes=module_names,
    module_name=module_name
))


def init(python_interpreter_path, max_pickle_version):
    # This function will look in the configurations directory and setup
    # the proper overrides
    config_dir = os.path.dirname(os.path.abspath(__file__)) + "/configurations"

    for path in os.listdir(config_dir):
        path = os.path.join(config_dir, path)
        if os.path.isdir(path):
            dir_name = os.path.basename(path)
            if dir_name.startswith('emulate_'):
                module_names = dir_name[8:].split("__")
                create_modules(
                    python_interpreter_path, max_pickle_version, path,
                    module_names)



import errno
import os
import json
import fcntl

CONDA_MAGIC_FILE = 'conda.dependencies'


def get_conda_manifest_path(ds_root, flow_name):
    return os.path.join(ds_root, flow_name, CONDA_MAGIC_FILE)


def read_conda_manifest(ds_root, flow_name):
    path = get_conda_manifest_path(ds_root, flow_name)
    if os.path.exists(path) and os.path.getsize(path) > 0:
        with open(path) as f:
            return json.load(f)
    else:
        return {}


def write_to_conda_manifest(ds_root, flow_name, key, value):
    path = get_conda_manifest_path(ds_root, flow_name)
    try:
        os.makedirs(os.path.dirname(path))
    except OSError as x:
        if x.errno != errno.EEXIST:
            raise
    with os.fdopen(os.open(path, os.O_RDWR | os.O_CREAT), 'r+') as f:
        try:
            fcntl.flock(f, fcntl.LOCK_EX)
            data = {}
            if os.path.getsize(path) > 0:
                f.seek(0)
                data = json.load(f)
            data[key] = value
            f.seek(0)
            json.dump(data, f)
            f.truncate()
        except IOError as e:
            if e.errno != errno.EAGAIN:
                raise
        finally:
            fcntl.flock(f, fcntl.LOCK_UN)


"""
S3 storage

Store data in S3
"""
import os
import json
import gzip
from io import BytesIO

try:
    # python2
    from urlparse import urlparse
except:
    # python3
    from urllib.parse import urlparse

from .. import metaflow_config
from .datastore import MetaflowDataStore, only_if_not_done
from ..metadata import MetaDatum
from .util.s3util import aws_retry, get_s3_client
from metaflow.util import Path

# We need UncloseableBytesIO for put_s3_object which may need
# to consume a BytesIO buffer multiple times. Blocking close()
# is cheaper than creating a new BytesIO object every time
# which would create duplicate copies of data.
class UncloseableBytesIO(BytesIO):
    def close(self):
        pass

class S3DataStore(MetaflowDataStore):
    TYPE='s3'

    s3 = None
    ClientError = None

    def __init__(self, *args, **kwargs):
        from .. import S3
        self.S3 = S3
        self.reset_client()
        super(S3DataStore, self).__init__(*args, **kwargs)

    @classmethod
    def reset_client(cls, hard_reset=False):
        # the s3 client is shared across all S3DataStores
        # so we don't open N connections to S3 unnecessarily
        if cls.s3 is None or hard_reset:
            cls.s3, cls.ClientError = get_s3_client()

    @aws_retry
    def _get_s3_object(self, path, return_buf=False):
        url = urlparse(path)
        buf = BytesIO()
        if self.monitor:
            with self.monitor.measure("metaflow.s3.get_object"):
                self.s3.download_fileobj(url.netloc, url.path.lstrip('/'), buf)
        else:
            self.s3.download_fileobj(url.netloc, url.path.lstrip('/'), buf)
        if return_buf:
            buf.seek(0)
            return buf
        else:
            return buf.getvalue()

    @aws_retry
    def _put_s3_object(self, path, blob=None, buf=None):
        url = urlparse(path)
        # @aws_retry may cause this function to be called multiple times with the same arguments.
        # Make sure that the buffer state is reset for every iteration
        if buf is None:
            buf = BytesIO(blob)
        else:
            buf.seek(0)
        if self.monitor:
            with self.monitor.measure("metaflow.s3.put_object"):
                self.s3.upload_fileobj(buf, url.netloc, url.path.lstrip('/'))
        else:
            self.s3.upload_fileobj(buf, url.netloc, url.path.lstrip('/'))

    @aws_retry
    def _head_s3_object(self, path):
        url = urlparse(path)
        try:
            return self.s3.head_object(Bucket=url.netloc, Key=url.path.lstrip('/'))
        except self.ClientError as err:
            error_code = int(err.response['Error']['Code'])
            if error_code == 404:
                return None
            else:
                raise

    @classmethod
    def get_latest_tasks(cls,
                         flow_name,
                         run_id=None,
                         steps=None,
                         pathspecs=None):
        run_prefix = cls.make_path(flow_name, run_id)

        from metaflow import S3
        with S3() as s3:
            task_urls = []
            # Note: When `pathspecs` is specified, we avoid the eventually
            # consistent `s3.list_paths` operation, and directly construct the
            # task_urls list.
            if pathspecs:
                task_urls = [cls.make_path(flow_name, pathspec=pathspec)
                             for pathspec in pathspecs]
            elif steps:
                task_objs = s3.list_paths(
                    [cls.make_path(flow_name, run_id, step) for step in steps])
                task_urls = [task.url for task in task_objs]
            else:
                step_objs = s3.list_paths([run_prefix])
                task_objs = s3.list_paths([step.url for step in step_objs])
                task_urls = [task.url for task in task_objs]
            urls = []
            for task_url in task_urls:
                for attempt in range(metaflow_config.MAX_ATTEMPTS):
                    metadata_filename = \
                        cls.get_metadata_filename_for_attempt(attempt)
                    urls.append(os.path.join(task_url, metadata_filename))
                    # Note for potential future optimization:
                    # Find the list of latest attempt for each task first, and
                    # follow up with a call to get done and metadata.
                    attempt_filename = \
                        cls.get_filename_for_attempt(attempt)
                    urls.append(os.path.join(task_url, attempt_filename))
                    done_filename = cls.get_done_filename_for_attempt(attempt)
                    urls.append(os.path.join(task_url, done_filename))

            results = s3.get_many(urls, return_missing=True)

            all_data_blobs = {}
            latest_attempt = {}
            done_attempts = set()

            for result in results:
                if result.exists:
                    path = result.url
                    step_name, task_id, fname = path.split('/')[-3:]
                    _, attempt = cls.parse_filename(fname)
                    if cls.is_done_filename(fname):
                        done_attempts.add((step_name, task_id, attempt))
                    elif cls.is_attempt_filename(fname):
                        # files are in sorted order, so overwrite is ok.
                        latest_attempt[(step_name, task_id)] = attempt
                    else:
                        # is_metadata_filename(fname) == True.
                        all_data_blobs[(step_name, task_id, attempt)] = \
                            result.blob
            latest_attempts = set((step_name, task_id, attempt)
                                   for (step_name, task_id), attempt
                                       in latest_attempt.items())
            latest_and_done = latest_attempts & done_attempts

        return [(step_name, task_id, attempt,
                 all_data_blobs[(step_name, task_id, attempt)])
                 for step_name, task_id, attempt in latest_and_done]

    @classmethod
    def get_artifacts(cls, artifacts_to_prefetch):
        artifact_list = []
        from metaflow import S3
        with S3() as s3:
            for obj in s3.get_many(artifacts_to_prefetch):
                sha = obj.key.split('/')[-1]
                artifact_list.append((sha, cls.decode_gzip_data(obj.path)))
        return artifact_list

    @only_if_not_done
    def save_metadata(self, name, data):
        """
        Save a task-specific metadata dictionary as JSON.
        """
        filename = self.filename_with_attempt_prefix('%s.json' % name,
                                                     self.attempt)
        path = os.path.join(self.root, filename)
        self._put_s3_object(path, json.dumps(data).encode('utf-8'))

    def has_metadata(self, name, with_attempt=True):
        attempt = self.attempt if with_attempt else None
        filename = self.filename_with_attempt_prefix('%s.json' % name,
                                                     attempt)
        path = os.path.join(self.root, filename)
        return bool(self._head_s3_object(path))

    def load_metadata(self, name):
        """
        Load a task-specific metadata dictionary as JSON.
        """
        filename = self.filename_with_attempt_prefix('%s.json' % name,
                                                     self.attempt)
        path = os.path.join(self.root, filename)
        return json.loads(self._get_s3_object(path).decode('utf-8'))

    def object_path(self, sha):
        root = os.path.join(self.data_root, sha[:2])
        return os.path.join(root, sha)

    @only_if_not_done
    def save_data(self, sha, transformable_object):
        """
        Save a content-addressed data blob if it doesn't exist already.
        """
        path = self.object_path(sha)
        if not self._head_s3_object(path):
            # we need UncloseableBytesIO for put_s3_object which may need
            # to consume the buffer multiple times
            buf = UncloseableBytesIO()
            # NOTE compresslevel makes a huge difference. The default
            # level of 9 can be impossibly slow.
            with gzip.GzipFile(fileobj=buf,
                               mode='wb',
                               compresslevel=3) as f:
                f.write(transformable_object.current())
            transformable_object.transform(lambda _: buf)
            buf.seek(0)
            self._put_s3_object(path, buf=buf)
        return path

    def load_data(self, sha):
        """
        Load a content-addressed data blob.
        """
        path = self.object_path(sha)
        buf = self._get_s3_object(path, return_buf=True)
        return self.decode_gzip_data(None, buf) # filename=None

    @only_if_not_done
    def save_logs(self, logsource, stream_data):
        """
        Save log files for multiple streams, represented as
        as a list of (stream, bytes) or (stream, Path) tuples.
        """
        locs = [self.get_log_location(logsource, stream)
                for stream, _ in stream_data]

        if stream_data and isinstance(stream_data[0][1], Path):
            keys = list(zip(locs, (str(path) for _, path in stream_data)))
            with self.S3() as s3:
                s3.put_files(keys)
        else:
            with self.S3() as s3:
                for url, data in zip(locs, (data for _, data in stream_data)):
                    s3.put(url, data)

    def load_log_legacy(self, stream, attempt_override=None):
        """
        Load old-style, pre-mflog, log file represented as a bytes object.
        """
        f = self.filename_with_attempt_prefix('%s.log' % stream,
                attempt_override if attempt_override is not None 
                    else self.attempt)
        path = os.path.join(self.root, f)
        if self._head_s3_object(path):
            return self._get_s3_object(path)
        else:
            return b''

    def load_logs(self, logsources, stream, attempt_override=None):
        urls = [self.get_log_location(source, stream, attempt_override) 
                    for source in logsources]
        with self.S3() as s3:
            results = s3.get_many(urls, return_missing=True)
            blobs = [log.blob if log.exists else b'' for log in results]
            return list(zip(logsources, blobs))


    def load_log(self, logtype, attempt_override=None):
        """
        Load a task-specific log file represented as a bytes object.
        """
        path = self.get_log_location(logtype, attempt_override)
        return self._get_s3_object(path)

    @only_if_not_done
    def done(self):
        """
        Write a marker indicating that datastore has finished writing to
        this path.
        """
        filename = self.get_done_filename_for_attempt(self.attempt)
        path = os.path.join(self.root, filename)
        self._put_s3_object(path, b'')

        self.metadata.register_metadata(
            self.run_id, self.step_name, self.task_id,
            [MetaDatum(field='attempt-done', value=str(self.attempt), type='attempt-done', tags=[])])

        self._is_done_set = True

    def is_done(self):
        """
        A flag indicating whether this datastore directory was closed
        succesfully with done().
        """
        filename = self.get_done_filename_for_attempt(self.attempt)
        path = os.path.join(self.root, filename)
        return bool(self._head_s3_object(path))



import gzip
import os
import sys
import time
from hashlib import sha1
from functools import partial

try:
    # Python 2
    import cPickle as pickle
except:
    # Python 3
    import pickle

from types import MethodType, FunctionType
from ..event_logger import NullEventLogger
from ..monitor import NullMonitor
from ..parameters import Parameter
from ..exception import MetaflowException, MetaflowInternalError
from ..metadata import DataArtifact
from .. import metaflow_config

class DataException(MetaflowException):
    headline = "Data store error"

class Inputs(object):
    """
    split-and: inputs.step_a.x inputs.step_b.x
    foreach: inputs[0].x
    both: (inp.x for inp in inputs)
    """
    def __init__(self, flows):
        # TODO sort by foreach index
        self.flows = list(flows)
        for flow in self.flows:
            setattr(self, flow._current_step, flow)

    def __getitem__(self, idx):
        return self.flows[idx]

    def __iter__(self):
        return iter(self.flows)

def only_if_not_done(f):
    def method(self, *args, **kwargs):
        if self._is_done_set:
            raise MetaflowInternalError("Tried to write to datastore "\
                                        "(method %s) after it was marked "\
                                        ".done()" % f.func_name)
        return f(self, *args, **kwargs)
    return method


class TransformableObject(object):
    # Very simple wrapper class to only keep one transform
    # of an object. This is to force garbage collection
    # on the transformed object if the transformation is
    # successful
    def __init__(self, current_object):
        self._object = current_object
        self._original_type = type(self._object)

    def transform(self, transformer):
        # Transformer is a function taking one argument (the current object) and returning another
        # object which will replace the current object if transformer does not raise an
        # exception
        try:
            temp = transformer(self._object)
            self._object = temp
        except:  # noqa E722
            raise

    def current(self):
        return self._object

    def current_type(self):
        return type(self._object)

    def original_type(self):
        return self._original_type


class MetaflowDataStore(object):
    datastore_root = None

    # Datastore needs to implement the methods below
    def save_metadata(self, name, data):
        """
        Save a task-specific metadata dictionary as JSON.
        """
        raise NotImplementedError()

    def load_metadata(self, name):
        """
        Load a task-specific metadata dictionary as JSON.
        """
        raise NotImplementedError()

    def has_metadata(self, name):
        """
        Return True if this metadata file exists.
        """
        raise NotImplementedError()

    def save_data(self, sha, transformable_object):
        """
        Save a content-addressed data blob if it doesn't exist already.
        """
        raise NotImplementedError()

    def load_data(self, sha):
        """
        Load a content-addressed data blob.
        """
        raise NotImplementedError()

    def save_logs(self, logsource, stream_data):
        """
        Save log files for multiple streams, represented as
        as a list of (stream, bytes) or (stream, Path) tuples.
        """
        raise NotImplementedError()

    def load_log_legacy(self, stream, attempt_override=None):
        """
        Load old-style, pre-mflog, log file represented as a bytes object.
        """
        raise NotImplementedError()

    def load_logs(self, logsources, stream, attempt_override=None):
        """
        Given a list of logsources, return a list of (logsource, logblob)
        tuples. Returns empty contents for missing log files.
        """
        raise NotImplementedError()

    def done(self):
        """
        Write a marker indicating that datastore has finished writing to
        this path.
        """
        raise NotImplementedError()

    def is_done(self):
        """
        A flag indicating whether this datastore directory was closed
        succesfully with done().
        """
        raise NotImplementedError()

    def object_path(self, sha):
        """
        Return URL of an object identified by a sha.
        """
        raise NotImplementedError()

    @classmethod
    def get_latest_tasks(cls,
                         flow_name,
                         run_id=None,
                         steps=None,
                         pathspecs=None):
        """
        Return a list of (step, task, attempt, metadata_blob) for a subset of
        the tasks (consider eventual consistency) for which the latest attempt
        is done for the input `flow_name, run_id`.
        We filter the list based on `steps` if non-None.
        Alternatively, `pathspecs` can contain the exact list of pathspec(s)
        (run_id/step_name/task_id) that should be filtered.
        Note: When `pathspecs` is specified, we expect strict consistency and
        not eventual consistency in contrast to other modes.
        """
        raise NotImplementedError()

    @classmethod
    def get_artifacts(cls, artifacts_to_prefetch):
        """
        Return a list of (sha, obj_blob) for all the object_path(s) specified in
        `artifacts_to_prefetch`.
        """
        raise NotImplementedError()

    def artifact_path(self, artifact_name):
        """
        Return the object_path() for `artifact_name`.
        Pre-condition: `artifact_name` is in datastore.
        """
        if self.objects:
            return self.object_path(self.objects[artifact_name])
        return None

    def get_log_location(self, logsource, stream, attempt_override=None):
        """
        Returns a string indicating the location of the log of the specified type.
        """
        filename = self.get_log_location_for_attempt(logsource, stream,
                        attempt_override if attempt_override is not None 
                            else self.attempt)
        return os.path.join(self.root, filename)

    @classmethod
    def get_datastore_root_from_config(cls, echo, create_on_absent=True):
        """
        Returns a default choice for datastore_root from metaflow_config
        depending on the datastore type.
        """
        return getattr(metaflow_config,
                       'DATASTORE_SYSROOT_%s' % cls.TYPE.upper())

    @classmethod
    def decode_gzip_data(cls, filename, fileobj=None):
        """
        Returns the gzip data in file with `filename` or passed via `fileobj`.
        """
        with gzip.GzipFile(filename, fileobj=fileobj, mode='rb') as f:
            return f.read()

    @classmethod
    def make_path(cls,
                  flow_name,
                  run_id=None,
                  step_name=None,
                  task_id=None,
                  pathspec=None):
        """
        Return the path for a given flow using this datastore.
        path = cls.datastore_root/flow_name/run_id/step_name/task_id
        Callers are expected to invoke this function with a sequence of non-None
        values depending on the nested path they want.
        For example,
        If run_id is None, return path = cls.datastore_root/flow_name
        If step_name is None, return path = cls.datastore_root/flow_name/run_id
        and so on.
        If pathspec is non None,
        return path = cls.datastore_root/flow_name/pathspec
        """
        sysroot = cls.datastore_root
        if pathspec:
            return os.path.join(sysroot, flow_name, pathspec)
        elif flow_name is None:
            return sysroot
        elif run_id is None:
            return os.path.join(sysroot, flow_name)
        elif step_name is None:
            return os.path.join(sysroot, flow_name, run_id)
        elif task_id is None:
            return os.path.join(sysroot, flow_name, run_id, step_name)
        else:
            return os.path.join(sysroot, flow_name, run_id, step_name, task_id)

    @classmethod
    def filename_with_attempt_prefix(cls, name, attempt):
        """
        Return the equivalent filename for `name` depending
        whether an attempt prefix must be used, if `attempt` isn't None.
        """
        if attempt is None:
            return name
        else:
            return '%d.%s' % (attempt, name)

    @classmethod
    def get_metadata_filename_for_attempt(cls, attempt):
        """
        Return the metadata filename (.data.json) based on `attempt`.
        """
        return cls.filename_with_attempt_prefix('data.json', attempt)

    @classmethod
    def get_log_location_for_attempt(cls, logprefix, stream, attempt):
        """
        Return the log_location based on `logprefix`, `stream` and `attempt`.
        """
        fname = '%s_%s.log' % (logprefix, stream)
        return cls.filename_with_attempt_prefix(fname, attempt)

    @classmethod
    def is_metadata_filename(cls, fname):
        """
        Returns if the filename is a metadata filename (ends in .data.json).
        """
        return fname.endswith('data.json')

    @classmethod
    def get_done_filename_for_attempt(cls, attempt):
        """
        Returns the done filename (.DONE.lock) based on `attempt`.
        """
        return cls.filename_with_attempt_prefix('DONE.lock', attempt)

    @classmethod
    def is_done_filename(cls, fname):
        """
        Returns if the filename is a done filename (ends in .DONE.lock).
        """
        return fname.endswith('DONE.lock')

    @classmethod
    def get_filename_for_attempt(cls, attempt):
        """
        Returns the attempt filename (.attempt.json) based on `attempt`.
        """
        return cls.filename_with_attempt_prefix('attempt.json', attempt)

    @classmethod
    def is_attempt_filename(cls, fname):
        """
        Returns if the filename is an attempt filename (ends in .attempt.json).
        """
        return fname.endswith('attempt.json')

    @classmethod
    def parse_filename(cls, fname):
        """
        Parse the filename and returns (name, attempt).
        When using old style paths (pre-attempt id), returns
        (name=fname, attempt=None).
        This is expected to be the converse to
        filename_with_attempt_prefix() method.
        """
        if len(fname) >= 1 and fname[0] >= '0' and fname[0] <= '9':
            # new style paths = <attempt>.<name>
            attempt = int(fname[0])
            name = fname[2:]
        else:
            # old style paths.
            attempt = None
            name = fname
        return name, attempt

    def __init__(self,
                 flow_name,
                 run_id=None,
                 step_name=None,
                 task_id=None,
                 mode='r',
                 metadata=None,
                 attempt=None,
                 event_logger=None,
                 monitor=None,
                 data_obj=None,
                 artifact_cache=None,
                 allow_unsuccessful=False):
        if run_id == 'data':
            raise DataException("Run ID 'data' is reserved. "
                                "Try with a different --run-id.")
        if self.datastore_root is None:
            raise DataException("Datastore root not found. "
                                "Specify with METAFLOW_DATASTORE_SYSROOT_%s "
                                "environment variable." % self.TYPE.upper())
        # NOTE: calling __init__(mode='w') should be a cheap operation:
        # no file system accesses are allowed. It is called frequently
        # e.g. to resolve log file location.
        self.event_logger = event_logger if event_logger else NullEventLogger()
        self.monitor = monitor if monitor else NullMonitor()
        self.metadata = metadata
        self.run_id = run_id
        self.step_name = step_name
        self.task_id = task_id
        self._is_done_set = False

        self._encodings = {'gzip+pickle-v2'}
        ver = sys.version_info[0] * 10 + sys.version_info[1]
        if ver >= 34:
            self._encodings.add('gzip+pickle-v4')

        self.artifact_cache = artifact_cache
        if self.artifact_cache is None:
            self.artifact_cache = {}

        self.data_root = os.path.join(self.datastore_root, flow_name, 'data')
        self.root = self.make_path(flow_name,
                                   run_id,
                                   step_name,
                                   task_id)

        self.attempt = attempt
        if mode == 'r':
            if data_obj is None:
                # what is the latest attempt ID of this data store?

                # In the case of S3, the has_metadata() below makes a
                # HEAD request to a non-existent object, which results
                # to this object becoming eventually consistent. This
                # could result to a situation that has_metadata() misses
                # the latest version although it is already existing.

                # As long as nothing opens a datastore for reading before
                # writing, this should not be a problem.

                # We have to make MAX_ATTEMPTS HEAD requests, which is
                # very unfortunate performance-wise (TODO: parallelize this).
                # On AWS Step Functions it is possible that some attempts are 
                # missing, so we have to check all possible attempt files to 
                # find the latest one. Compared to doing a LIST operation, 
                # these checks are guaranteed to be consistent as long as the 
                # task to be looked up has already finished.
                self.attempt = None # backwards-compatibility for pre-attempts.
                for i in range(0, metaflow_config.MAX_ATTEMPTS):
                    if self.has_metadata('%d.attempt' % i, with_attempt=False):
                        self.attempt = i

                # was the latest attempt completed successfully?
                if self.is_done():
                    # load the data from the latest attempt
                    data_obj = self.load_metadata('data')
                elif allow_unsuccessful and self.attempt is not None:
                    # this mode can be used to load_logs, for instance
                    data_obj = None
                else:
                    raise DataException("Data was not found or not finished at %s"\
                                        % self.root)

            if data_obj:
                self.origin = data_obj.get('origin')
                self.objects = data_obj['objects']
                self.info = data_obj.get('info', {})
        elif mode == 'd':
            # Direct access mode used by the client. We effectively don't load any
            # objects and can only access things using the load_* functions
            self.origin = None
            self.objects = None
            self.info = None
        elif mode != 'w':
            raise DataException('Unknown datastore mode: %s' % mode)

    def init_task(self):
        # this method should be called once after datastore has been opened
        # for task-related write operations
        self.save_metadata('attempt', {'time': time.time()})
        self.objects = {}
        self.info = {}


    @property
    def pathspec(self):
        return '%s/%s/%s' % (self.run_id, self.step_name, self.task_id)

    @property
    def pathspec_index(self):
        idxstr = ','.join(map(str, (f.index for f in self['_foreach_stack'])))
        return '%s/%s[%s]' % (self.run_id, self.step_name, idxstr)

    def _save_object(self, transformable_obj, var, force_v4=False):
        if force_v4:
            blobtype = 'gzip+pickle-v4'
            if blobtype not in self._encodings:
                raise DataException("Artifact *%s* requires a serialization encoding that "
                                    "requires Python 3.4 or newer." % var)
            transformable_obj.transform(lambda x: pickle.dumps(x, protocol=4))
        else:
            try:
                # to ensure compatibility between python2 and python3, we use the
                # highest protocol that works with both the versions
                transformable_obj.transform(lambda x: pickle.dumps(x, protocol=2))
                blobtype = 'gzip+pickle-v2'
            except (SystemError, OverflowError):
                # this happens when you try to serialize an oversized
                # object (2GB/4GB+)
                blobtype = 'gzip+pickle-v4'
                if blobtype not in self._encodings:
                    raise DataException("Artifact *%s* is very large (over 2GB). "
                                        "You need to use Python 3.4 or newer if "
                                        "you want to serialize large objects."
                                        % var)
                transformable_obj.transform(lambda x: pickle.dumps(x, protocol=4))
        sha = sha1(transformable_obj.current()).hexdigest()
        sz = len(transformable_obj.current())
        self.save_data(sha, transformable_obj)
        return sha, sz, blobtype

    def _load_object(self, sha):
        if sha in self.artifact_cache:
            blob = self.artifact_cache[sha]
        else:
            blob = self.load_data(sha)
        return pickle.loads(blob)

    @only_if_not_done
    def clone(self, origin):
        self.save_metadata('data', {'datastore': self.TYPE,
                                    'version': '0.1',
                                    'origin': origin.pathspec,
                                    'python_version': sys.version,
                                    'objects': origin.objects,
                                    'info': origin.info})
        self._register_data_artifacts(origin.objects, origin.info)

    @only_if_not_done
    def passdown_partial(self, origin, vars):
        # Pass-down from datastore origin all information related to vars to
        # this datastore. In other words, this adds to the current datastore all
        # the variables in vars (obviously, it does not download them or anything but
        # records information about them). This is used to propagate parameters between
        # datastores without actually loading the parameters
        for var in vars:
            sha = origin.objects.get(var)
            if sha:
                self.objects[var] = sha
                self.info[var] = origin.info[var]

    @only_if_not_done
    def persist(self, flow):

        def serializable_attributes():
            for var in dir(flow):
                if var.startswith('__') or var in flow._EPHEMERAL:
                    continue
                # Skip over properties of the class (Parameters)
                if hasattr(flow.__class__, var) and \
                        isinstance(getattr(flow.__class__, var), property):
                    continue
                val = getattr(flow, var)
                if not (isinstance(val, MethodType) or
                        isinstance(val, FunctionType) or
                        isinstance(val, Parameter)):
                    yield var, TransformableObject(val), False

        # initialize with old values...
        if flow._datastore:
            self.objects.update(flow._datastore.objects)
            self.info.update(flow._datastore.info)

        # ...overwrite with new
        for var, obj, force_v4 in serializable_attributes():
            sha, size, encoding = self._save_object(obj, var, force_v4)
            self.objects[var] = sha
            self.info[var] = {'size': size,
                              'type': str(obj.original_type()),
                              'encoding': encoding}

        self.save_metadata('data', {'datastore': self.TYPE,
                                    'version': '1.0',
                                    'attempt': self.attempt,
                                    'python_version': sys.version,
                                    'objects': self.objects,
                                    'info': self.info})

        self._register_data_artifacts(self.objects, self.info)

    def _register_data_artifacts(self, objects, info):
        # register artifacts with the metadata service
        artifacts = [DataArtifact(name=var,
                                  ds_type=self.TYPE,
                                  url=self.object_path(sha),
                                  sha=sha,
                                  type=info[var]['encoding'])
                     for var, sha in objects.items()]

        self.metadata.register_data_artifacts(self.run_id,
                                              self.step_name,
                                              self.task_id,
                                              self.attempt,
                                              artifacts)

    def get(self, var, default=None):
        if self.objects:
            return self[var] if var in self.objects else default
        return default

    # Provides a fast-path to check if a given object is None.
    def is_none(self, var):
        if not self.info:
            return True
        info = self.info.get(var)
        if info:
            obj_type = info.get('type')
            # Conservatively check if the actual object is None, in case
            # the artifact is stored using a different python version.
            if obj_type == str(type(None)):
                return True
        # Slow path since this has to get the object from S3.
        return self.get(var) is None

    def __contains__(self, var):
        if self.objects:
            return var in self.objects
        return False

    def __getitem__(self, var):
        # backwards compatibility: we might not have info for all objects
        if not self.info:
            return None
        info = self.info.get(var)
        if info:
            encoding = info.get('encoding', 'gzip+pickle-v2')
        else:
            encoding = 'gzip+pickle-v2'
        if encoding in self._encodings:
            return self._load_object(self.objects[var])
        raise DataException("Artifact *%s* requires a newer version "
                            "of Python. Try with Python 3.4 or newer." %
                            var)

    def __iter__(self):
        if self.objects:
            return iter(self.objects)
        return iter([])

    def __str__(self):
        return self.format(show_private=True, max_value_size=1000)

    def items(self):
        if self.objects:
            return self.objects.items()
        return {}

    def to_dict(self, show_private=False, max_value_size=None, include=None):
        d = {}
        for k, v in self.items():
            if include and k not in include:
                continue
            if k[0] == '_' and not show_private:
                continue
            if max_value_size is not None and\
               self.info[k]['size'] > max_value_size:
                d[k] = ArtifactTooLarge()
            else:
                d[k] = self[k]
        return d

    def format(self, **kwargs):
        def lines():
            for k, v in self.to_dict(**kwargs).items():
                yield k, '*{key}* [size: {size} type: {type}] = {value}'\
                         .format(key=k, value=v, **self.info[k])
        return '\n'.join(line for k, line in sorted(lines()))

class ArtifactTooLarge(object):
    def __str__(self):
        return '< artifact too large >'



import traceback

from metaflow.exception import MetaflowException,\
    MetaflowExceptionWrapper
from metaflow.decorators import StepDecorator
from metaflow.unbounded_foreach import UBF_CONTROL

NUM_FALLBACK_RETRIES = 3


class FailureHandledByCatch(MetaflowException):
    headline = 'Task execution failed but @catch handled it'

    def __init__(self, retry_count):
        msg = 'Task execution kept failing over %d attempts. '\
              'Your code did not raise an exception. Something '\
              'in the execution environment caused the failure.' % retry_count
        super(FailureHandledByCatch, self).__init__(msg)


class CatchDecorator(StepDecorator):
    """
    Step decorator to specify error handling for your step.

    This decorator indicates that exceptions in the step should be caught and not fail the entire
    flow.

    This can be used in conjunction with the @retry decorator. In that case, catch will only
    activate if all retries fail and will catch the last exception thrown by the last retry.

    To use, annotate your step as follows:
    ```
    @catch(var='foo')
    @step
    def myStep(self):
        ...
    ```

    Parameters
    ----------
    var : string
        Name of the artifact in which to store the caught exception. If not specified,
        the exception is not stored
    print_exception : bool
        Determines whether or not the exception is printed to stdout when caught. Defaults
        to True
    """
    name = 'catch'
    defaults = {'var': None,
                'print_exception': True}

    def step_init(self, flow, graph, step, decos, environment, datastore, logger):
        # handling _foreach_var and _foreach_num_splits requires some
        # deeper thinking, so let's not support that use case for now
        self.logger = logger
        if graph[step].type == 'foreach':
            raise MetaflowException('@catch is defined for the step *%s* '
                                    'but @catch is not supported in foreach '
                                    'split steps.' % step)

    def _print_exception(self, step, flow):
        self.logger(head='@catch caught an exception from %s' % flow,
                    timestamp=False)
        for line in traceback.format_exc().splitlines():
            self.logger('>  %s' % line, timestamp=False)

    def _set_var(self, flow, val):
        var = self.attributes.get('var')
        if var:
            setattr(flow, var, val)

    def task_exception(self,
                       exception,
                       step,
                       flow,
                       graph,
                       retry_count,
                       max_user_code_retries):

        if self.attributes['print_exception']:
            self._print_exception(step, flow)

        # pretend that self.next() was called as usual
        flow._transition = (graph[step].out_funcs, None, None)
        # store the exception
        picklable = MetaflowExceptionWrapper(exception)
        flow._catch_exception = picklable
        self._set_var(flow, picklable)
        return True

    def task_post_step(self,
                       step_name,
                       flow,
                       graph,
                       retry_count,
                       max_user_code_retries):
        # there was no exception, set the exception var (if any) to None
        self._set_var(flow, None)

    def step_task_retry_count(self):
        return 0, NUM_FALLBACK_RETRIES

    def task_decorate(self,
                      step_func,
                      func,
                      graph,
                      retry_count,
                      max_user_code_retries,
                      ubf_context):

        # if the user code has failed max_user_code_retries times, @catch
        # runs a piece of fallback code instead. This way we can continue
        # running the flow downsteam, as we have a proper entry for this task.

        def fallback_step(inputs=None):
            raise FailureHandledByCatch(retry_count)

        # We don't run fallback for `ubf_control` since it doesn't support
        # any retries.
        if ubf_context != UBF_CONTROL and retry_count > max_user_code_retries:
            return fallback_step
        else:
            return step_func
